import os
from llama_index.core.indices.query.query_transform import HyDEQueryTransform
from llama_index.llms.openai import OpenAI
from llama_index.core import PromptTemplate
from llama_index.core.prompts.prompt_type import PromptType
from llama_index.core.tools import QueryEngineTool, ToolMetadata
from llama_index.core.query_engine import SubQuestionQueryEngine
from llama_index.question_gen.openai import OpenAIQuestionGenerator
from llama_index.core.schema import NodeWithScore, TextNode

from ..llms import llm

# 定义模板字符串
ZEROSHOT_OPENAI_SUB_QUESTION_PROMPT_TMPL = """\
You are a world class state of the art agent.

You have access to multiple tools, each representing a different data source or API.
Each of the tools has a name and a description, formatted as a JSON dictionary.
The keys of the dictionary are the names of the tools and the values are the \
descriptions.
Your purpose is to help answer a complex user question by generating a list of sub \
questions that can be answered by the tools.

These are the guidelines you consider when completing your task:
* Be as specific as possible
* The sub questions should be relevant to the user question
* The sub questions should be answerable by the tools provided
* You can generate multiple sub questions for each tool
* Tools must be specified by their name, not their description
* You don't need to use a tool if you don't think it's relevant

Output the list of sub questions by calling the SubQuestionList function.

## Tools
```json
{tools_str}
```

## User Question
{query_str}
"""

FEWSHOT_OPENAI_SUB_QUESTION_PROMPT_TMPL = """\
You are a world class state of the art agent.

You have access to multiple tools, each representing a different data source or API.
Each of the tools has a name and a description, formatted as a JSON dictionary.
The keys of the dictionary are the names of the tools and the values are the \
descriptions.
Your purpose is to help answer a complex user question by generating a list of sub \
questions that can be answered by the tools.

These are the guidelines you consider when completing your task:
* Be as specific as possible
* The sub questions should be relevant to the user question
* The sub questions should be answerable by the tools provided
* You can generate multiple sub questions for each tool
* Tools must be specified by their name, not their description
* You don't need to use a tool if you don't think it's relevant

Here are a few examples:

1.
Question: What type of media does Tar Creek and Before Stonewall have in common?
Sub questions:
- What is the media type of Tar Creek?
- What is the media type of Before Stonewall?

2.
Question: Which documentary was created first, Murderball or Year at Danger?
Sub questions: 
- When was the documentary "Murderball" created?
- When was the documentary "Year at Danger" created?

3.
Question: How was Paul Grahams life different before, during, and after YC?
Sub questions:
- What did Paul Graham work on before YC?
- What did Paul Graham work on during YC?
- What did Paul Graham work on after YC?

4.
Question: What role did Ed Roland and Adam Duritz perform in their respective bands, Collective Soul and Counting Crows?
Sub questions: 
- What role did Ed Roland perform in the band Collective Soul?
- What role did Adam Duritz perform in the band Counting Crows?

5.
Question: Did Down and Out in America or Lost in La Mancha win more awards?
Sub questions: 
- How many awards did the movie "Down and Out in America" win?
- How many awards did the movie "Lost in La Mancha" win?

Output the list of sub questions by calling the SubQuestionList function.

## Tools
```json
{tools_str}
```

## User Question
{query_str}
"""

def transform_and_query(query, cfg, query_engine):
    """同步版本的查询转换函数"""
    if cfg.query_transform == "subquery_zeroshot":
        return subquery_zeroshot_sync(query, query_engine)
    elif cfg.query_transform == "subquery_fewshot":
        return subquery_fewshot_sync(query, query_engine)
    else:
        transformed_query = transform(llm, query, cfg)
        return query_engine.query(transformed_query)

async def transform_and_query_async(query, cfg, query_engine):
    """异步版本的查询转换函数"""
    if cfg.query_transform == "subquery_zeroshot":
        return await subquery_zeroshot(query, query_engine)
    elif cfg.query_transform == "subquery_fewshot":
        return await subquery_fewshot(query, query_engine)
    else:
        transformed_query = transform(llm, query, cfg)
        return await query_engine.aquery(transformed_query)


def transform(llm, query, cfg):
    # parser.add_argument('--query_transform', type=str, default='none', help='query transform, available: none, hyde_zeroshot, hyde_fewshot, stepback_zeroshot, stepback_fewshot, subquery_zeroshot, subquery_fewshot')
    if cfg.query_transform == "none":
        return query
    elif cfg.query_transform == "hyde_zeroshot":
        return hyde_zeroshot(query)
    elif cfg.query_transform == "hyde_fewshot":
        return hyde_fewshot(query)
    elif cfg.query_transform == "stepback_zeroshot":
        return stepback_zeroshot(query)
    elif cfg.query_transform == "stepback_fewshot":
        return stepback_fewshot(query)
    else:
        raise Exception("Unknown query transform: %s" % cfg.query_transform)


def hyde(query, prompt_template_str):
    """
    Query改写 - HyDE
    """
    HYDE_PROMPT = PromptTemplate(prompt_template_str, prompt_type=PromptType.SUMMARY)
    hyde = HyDEQueryTransform(include_original=True, hyde_prompt=HYDE_PROMPT)
    query_hyde = hyde.run(query)
    return (
            query_hyde.custom_embedding_strs[0] + " " + query_hyde.custom_embedding_strs[1]
    )


def hyde_zeroshot(query):
    """
    Query改写 - HyDE - zeroshot
    """
    HYDE_TMPL_ZEROSHOT = (
        "Please write a passage to answer the question\n"
        "Try to include as many key details as possible.\n"
        "\n"
        "Here is your task:"
        "Question: {context_str}\n"
        "Passage: "
    )

    return hyde(query, HYDE_TMPL_ZEROSHOT)


def hyde_fewshot(query):
    """
    Query改写 - HyDE - fewshot
    """
    HYDE_TMPL_FEWSHOT = (
        "Please write a passage to answer the question\n"
        "Try to include as many key details as possible.\n"
        "\n"
        "Here are a few examples:\n"
        "\n"
        "\n"
        'Passage:"""\n'
        "1.\n"
        "Question: What is Bel?\n"
        "Passage: Bel is a term that has multiple meanings and can be interpreted in various ways depending on the context. In ancient Mesopotamian mythology, Bel was a prominent deity and the god of the heavens and earth. He was considered the supreme god and was often associated with the city of Babylon. Bel was believed to have control over the forces of nature and was worshipped by the Babylonians through elaborate rituals and offerings.\n\n"
        "In addition to its mythological significance, Bel is also a title given to individuals who hold high positions of authority or leadership. For example, in the ancient Near East, the title of Bel was used to refer to the ruler or king of a city-state. This title denoted power, influence, and the ability to govern and make important decisions.\n\n"
        "Furthermore, Bel is also a term used in the field of linguistics to refer to a unit of sound or phoneme. In phonetics, a bel represents a logarithmic unit used to measure the intensity or loudness of sound. This unit is named after Alexander Graham Bell, the inventor of the telephone, who made significant contributions to the study of sound and communication.\n\n"
        "Overall, the term Bel encompasses a range of meanings and can refer to a mythical deity, a title of authority, or a unit of sound measurement. Its significance varies depending on the context in which it is used, highlighting the diverse nature of this term.\n"
        "2.\n"
        "Question: At It Again contains lyrics co-written by the singer and actor from what city?\n"
        "Passage: The song 'At It Again' features lyrics that were co-written by the singer and actor who hail from the vibrant city of Los Angeles. This city, known for its thriving entertainment industry, has been a hub for countless talented individuals who have made their mark in music, film, and television. The singer and actor, both born and raised in Los Angeles, have been deeply influenced by the diverse and creative atmosphere of their hometown. Their collaboration on 'At It Again' showcases their unique perspectives and storytelling abilities, as they draw inspiration from their personal experiences and the rich cultural tapestry of Los Angeles. Through their lyrics, they paint a vivid picture of the city's energy, its dreams, and its challenges, capturing the essence of their beloved hometown in every verse.\n"
        "3.\n"
        "Question: What year was the English actor and film producer born in who also starred in an adaptation with Renee Zellweger?\n"
        "Passage: The English actor and film producer who starred alongside Renee Zellweger in a film adaptation is Hugh Grant. Born on September 9, 1960, Grant has had a prolific career in the film industry, particularly known for his roles in romantic comedies. He is a native of Hammersmith, London, and he attended the New College at the University of Oxford, where he studied English literature.\n\n"
        "Grant's breakthrough role came in 1994 when he starred in 'Four Weddings and a Funeral.' His performance in this film garnered him international recognition and established his reputation as a leading man in romantic comedy films. His charm and unique comedic timing resonated with audiences worldwide, making him a favorite in the genre.\n\n"
        "In 2001, Grant starred in the film adaptation of 'Bridget Jones's Diary,' alongside American actress Renee Zellweger. The film, based on Helen Fielding's popular novel, was a commercial success, grossing over $280 million worldwide. Grant played the character of Daniel Cleaver, a charming yet unreliable publishing executive who becomes involved in a love triangle with Bridget Jones (Zellweger) and Mark Darcy (Colin Firth).\n\n"
        "'Bridget Jones's Diary' was well-received by critics, and it led to two sequels in which Grant reprised his role. His performance in these films is often considered one of his most memorable. Despite the various challenges in his career, Hugh Grant has remained a significant figure in the film industry for over three decades. His enduring appeal and talent have ensured his place in the annals of cinematic history.\n"
        "4.\n"
        "Question: What is the length of the track where the 2013 Liqui Moly Bathurst 12 Hour was staged?\n"
        "Passage: The 2013 Liqui Moly Bathurst 12 Hour race was held at the iconic Mount Panorama Circuit, situated in Bathurst, New South Wales, Australia. This particular motor racing track is known not only for its challenging layout but also for the spectacular views it offers of the surrounding countryside.\n\n"
        "The Mount Panorama Circuit is a unique blend of seemingly contrasting elements. It is comprised of both long and short straights, fast corners, as well as dramatic ascents and descents, which demand a high level of skill and precision from the drivers. The track's total length is approximately 6.213 kilometers, or about 3.861 miles, a distance that has remained consistent over the years.\n\n"
        "The Mount Panorama Circuit is steeped in motor racing history. It was first opened in 1938 and has since been the venue for a variety of prestigious motor racing events. These include not only the Bathurst 12 Hour, but also the Bathurst 1000, the latter of which is considered one of the most significant events in Australian motorsport.\n\n"
        "The 2013 Liqui Moly Bathurst 12 Hour race was a significant event in the racing calendar. Teams from around the world competed in a gruelling endurance race that tested not only the performance of the cars but also the stamina and skill of the drivers. The race, which started in the early morning darkness, ran for twelve hours, with the victors crossing the finish line as the sun was setting, adding to the drama and spectacle of the event.\n\n"
        "The Mount Panorama Circuit's unique combination of high-speed straights and tight corners, along with its significant changes in elevation, make it one of the most challenging and exciting circuits in the world. Its length and layout have remained largely unchanged since its inception, providing a consistent benchmark for drivers and teams across different racing events. It's a track that demands respect from every driver who races on it, and the 2013 Liqui Moly Bathurst 12 Hour was no exception.\n"
        "5.\n"
        "Question: Which event was a surprise military strike by the Imperial Japanese Navy Air Service while Magruder Hill Tuttle was a senior officer in the United States Navy?\n"
        "Passage: The event that was a surprise military strike by the Imperial Japanese Navy Air Service while Magruder Hill Tuttle was a senior officer in the United States Navy was the infamous attack on Pearl Harbor. On the morning of December 7, 1941, the peaceful tranquility of the Hawaiian naval base was shattered as a fleet of Japanese aircraft descended upon the unsuspecting American forces stationed there. Tuttle, who held a senior position within the US Navy, found himself thrust into the chaos and confusion of the surprise attack.\n\n"
        "The Japanese planes unleashed a devastating assault, targeting the American battleships, cruisers, and aircraft on the island. The attack was swift and relentless, catching the US Navy off guard and causing immense destruction and loss of life. Tuttle, along with his fellow officers and sailors, valiantly fought back against the overwhelming enemy forces, but the surprise nature of the attack made it difficult to mount an effective defense.\n\n"
        "The attack on Pearl Harbor marked a turning point in World War II, propelling the United States into the conflict and forever changing the course of history.\n"
        "\n"
        "Here is your task:"
        "Question: {context_str}\n"
        "Passage: "
    )

    return hyde(query, HYDE_TMPL_FEWSHOT)


def stepback(query, prompt_template_str, llm=llm):
    """
    Query扩写 - Stepback
    """
    stepback_query_gen_prompt = PromptTemplate(prompt_template_str)
    return llm.predict(stepback_query_gen_prompt, query=query) + " " + query


def stepback_zeroshot(query, llm=llm):
    """
    Query扩写 - Stepback - zeroshot
    """
    stepback_query_gen_str_zeroshot = """\
    You are an expert at world knowledge. Your task is to step back and paraphrase a question to a more generic step-back question, which is easier to answer.
    Just give the content of the Stepback Question.

    Here is your task:
    Original Question: {query}
    Stepback Question: 
    """
    return stepback(query, stepback_query_gen_str_zeroshot, llm=llm)


def stepback_fewshot(query, llm=llm):
    """
    Query拆解 - Stepback - fewshot
    """

    stepback_query_gen_str_fewshot = """\
    You are an expert at world knowledge. Your task is to step back and paraphrase a question to a more generic step-back question, which is easier to answer.
    Just give the content of the Stepback Question.

    Here are a few examples:

    1.
    Original Question: Musician and satirist Allie Goertz wrote a song about the "The Simpsons" character Milhouse, who Matt Groening named after who?
    Stepback Question: Who are some notable figures that have influenced the names of "The Simpsons" characters?

    2.
    Original Question: Which plant's name is derived from the greek for "bee," Melittis or Hovea?
    Stepback Question: What are the origins of the names for the plants Melittis and Hovea?

    3.
    Original Question: What American media franchise is centered on a series of monster films featuring Godzilla and King Kong?
    Stepback Question: What are some notable American media franchises that feature monster films?

    4.
    Original Question: Tar Creed Superfund site causing toxic areas and therefore reducing the population to what number in November 2010?
    Stepback Question: What impact has the Tar Creek Superfund site had on the population of nearby towns, such as Cardin, Oklahoma?

    5.
    Original Question: Who is likely to have represented and advanced Armenian music more, Art Laboe or Konstantin Orbelyan?
    Stepback Question: Who are some notable figures that have significantly contributed to the representation and advancement of Armenian music?
    
    
    Here is your task:
    Original Question: {query}
    Stepback Question: 
    """

    return stepback(query, stepback_query_gen_str_fewshot, llm=llm)


def subquery_sync(query, prompt_template_str, query_engine):
    """同步版本的 subquery 函数"""
    query_engine_tools = [
        QueryEngineTool(
            query_engine=query_engine,
            metadata=ToolMetadata(
                name="Data Source",
                description="Data source for the query engine",
            ),
        )
    ]

    subquery_engine = CustomSubQuestionQueryEngine.from_defaults(
        query_engine_tools=query_engine_tools,
        use_async=False,
        question_gen=OpenAIQuestionGenerator.from_defaults(
            prompt_template_str=prompt_template_str
        ),
    )

    return subquery_engine.query(query)

def subquery_zeroshot_sync(query, query_engine):
    """同步版本的 subquery_zeroshot 函数"""
    return subquery_sync(query, ZEROSHOT_OPENAI_SUB_QUESTION_PROMPT_TMPL, query_engine)

def subquery_fewshot_sync(query, query_engine):
    """同步版本的 subquery_fewshot 函数"""
    return subquery_sync(query, FEWSHOT_OPENAI_SUB_QUESTION_PROMPT_TMPL, query_engine)


class CustomSubQuestionQueryEngine(SubQuestionQueryEngine):
    def _construct_node(self, qa_pair):
        node_text = f"Sub question: {qa_pair.sub_q.sub_question}\nResponse: {qa_pair.answer}"
        if qa_pair.sources and len(qa_pair.sources) > 0:
            metadata = qa_pair.sources[0].node.metadata.copy()
        else:
            metadata = {}
        node = TextNode(text=node_text, metadata=metadata)
        return NodeWithScore(node=node)

async def subquery(query, prompt_template_str, query_engine):
    """异步版本的 subquery 函数"""
    query_engine_tools = [
        QueryEngineTool(
            query_engine=query_engine,
            metadata=ToolMetadata(
                name="Data Source",
                description="Data source for the query engine",
            ),
        )
    ]

    subquery_engine = CustomSubQuestionQueryEngine.from_defaults(
        query_engine_tools=query_engine_tools,
        use_async=True,
        question_gen=OpenAIQuestionGenerator.from_defaults(
            prompt_template_str=prompt_template_str
        ),
    )

    return await subquery_engine.aquery(query)

async def subquery_zeroshot(query, query_engine):
    """异步版本的 subquery_zeroshot 函数"""
    return await subquery(query, ZEROSHOT_OPENAI_SUB_QUESTION_PROMPT_TMPL, query_engine)

async def subquery_fewshot(query, query_engine):
    """异步版本的 subquery_fewshot 函数"""
    return await subquery(query, FEWSHOT_OPENAI_SUB_QUESTION_PROMPT_TMPL, query_engine)



from functools import partial

import torch
from transformers import AutoTokenizer, AutoModelForCausalLM, AutoModel
from transformers.generation.utils import GenerationConfig
from llama_index.llms.huggingface import HuggingFaceLLM
# pip install llama-index-llms-huggingface
from ..config import Config
cfg = Config()

load_tokenizer = []


def llama_model_and_tokenizer(name, auth_token):
    # Create tokenizer
    tokenizer = AutoTokenizer.from_pretrained(name, token=auth_token)

    # Create model
    model = AutoModelForCausalLM.from_pretrained(name, token=auth_token, torch_dtype=torch.float16,
                                                 rope_scaling={"type": "dynamic", "factor": 2},
                                                 load_in_8bit=True, device_map="auto").eval()

    return tokenizer, model


def llama_completion_to_prompt(completion):
    return f"""<s>[INST] <<SYS>>
        You are a helpful, respectful and honest assistant. Always answer as 
        helpfully as possible, while being safe. Your answers should not include
        any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content.
        Please ensure that your responses are socially unbiased and positive in nature.

        If a question does not make any sense, or is not factually coherent, explain 
        why instead of answering something not correct. If you don't know the answer 
        to a question, please don't share false information.

        Your goal is to provide answers relating to the financial performance of 
        the company.<</SYS>>
        {completion} [/INST]"""


def chatglm_model_and_tokenizer(name):
    # Create tokenizer
    tokenizer = AutoTokenizer.from_pretrained(name, trust_remote_code=True)
    load_tokenizer.append(tokenizer)

    # Create model
    model = AutoModel.from_pretrained(name, trust_remote_code=True).half().cuda().eval()

    return tokenizer, model


def chatglm_completion_to_prompt(completion):
    return "<|user|>\n " + completion + "<|assistant|>"


def qwen_model_and_tokenizer(name):
    # Create tokenizer
    tokenizer = AutoTokenizer.from_pretrained(name)
    load_tokenizer.append(tokenizer)

    # Create model
    model = AutoModelForCausalLM.from_pretrained(name,
                                                 torch_dtype=torch.float16,
                                                 device_map="auto").eval()

    return tokenizer, model


def qwen_completion_to_prompt(completion):
    tokenizer = load_tokenizer[0]
    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": completion}
    ]
    return tokenizer.apply_chat_template(
        messages,
        tokenize=False,
        add_generation_prompt=True
    )


def baichuan_model_and_tokenizer(name):
    # Create tokenizer
    tokenizer = AutoTokenizer.from_pretrained(name, use_fast=False, trust_remote_code=True)
    load_tokenizer.append(tokenizer)

    # Create model
    model = AutoModelForCausalLM.from_pretrained(name, torch_dtype=torch.float16, trust_remote_code=True, device_map="auto").eval()
    model.generation_config = GenerationConfig.from_pretrained("baichuan-inc/Baichuan2-7B-Chat")

    return tokenizer, model


def baichuan_completion_to_prompt(completion):
    return "<reserved_106>" + completion + "<reserved_107>"  # "You are a helpful assistant.<reserved_106>" + completion + "<reserved_107>""


def falcon_model_and_tokenizer(name):
    # Create tokenizer
    tokenizer = AutoTokenizer.from_pretrained(name)
    load_tokenizer.append(tokenizer)

    # Create model
    model = AutoModelForCausalLM.from_pretrained(name, trust_remote_code=True, device_map="auto").eval()

    return tokenizer, model


def falcon_completion_to_prompt(completion):
    return completion


def mpt_model_and_tokenizer(name):
    # Create tokenizer
    tokenizer = AutoTokenizer.from_pretrained("EleutherAI/gpt-neox-20b")
    load_tokenizer.append(tokenizer)

    # Create model
    model = AutoModelForCausalLM.from_pretrained(name, trust_remote_code=True, device_map="auto").eval()

    return tokenizer, model


def mpt_completion_to_prompt(completion):
    return completion


def yi_model_and_tokenizer(name):
    # Create tokenizer
    tokenizer = AutoTokenizer.from_pretrained(name, use_fast=False)
    load_tokenizer.append(tokenizer)

    # Create model
    model = AutoModelForCausalLM.from_pretrained(
        name,
        torch_dtype=torch.float16,
        device_map="auto"
    ).eval()

    return tokenizer, model


def yi_completion_to_prompt(completion):
    return "<|im_start|> user\n" + completion + "<|im_end|> \n<|im_start|>assistant\n"


tokenizer_and_model_fn_dict = {
    "meta-llama/Llama-2-7b-chat-hf": partial(llama_model_and_tokenizer, auth_token=cfg.auth_token),
    "THUDM/chatglm3-6b": chatglm_model_and_tokenizer,
    "Qwen/Qwen1.5-7B-Chat": qwen_model_and_tokenizer,
    "Qwen/Qwen1.5-7B-Chat-GPTQ-Int8": qwen_model_and_tokenizer,
    "baichuan-inc/Baichuan2-7B-Chat": baichuan_model_and_tokenizer,
    "tiiuae/falcon-7b-instruct": falcon_model_and_tokenizer,
    "mosaicml/mpt-7b-chat": mpt_model_and_tokenizer,
    "01-ai/Yi-6B-Chat": yi_model_and_tokenizer,
}

completion_to_prompt_dict = {
    "meta-llama/Llama-2-7b-chat-hf": llama_completion_to_prompt,
    "THUDM/chatglm3-6b": chatglm_completion_to_prompt,
    "Qwen/Qwen1.5-7B-Chat": qwen_completion_to_prompt,
    "Qwen/Qwen1.5-7B-Chat-GPTQ-Int8": qwen_completion_to_prompt,
    "baichuan-inc/Baichuan2-7B-Chat": baichuan_completion_to_prompt,
    "tiiuae/falcon-7b-instruct": falcon_completion_to_prompt,
    "mosaicml/mpt-7b-chat": mpt_completion_to_prompt,
    "01-ai/Yi-6B-Chat": yi_completion_to_prompt,
}

# llm_argument_dict = {
#     "meta-llama/Llama-2-7b-chat-hf": {"context_window": 4096, "max_new_tokens": 256,
#                                       "generate_kwargs": {"temperature": 0.7, "top_k": 50, "top_p": 0.95}},
#     "THUDM/chatglm3-6b": {"context_window": 4096, "max_new_tokens": 256,
#                           "generate_kwargs": {"do_sample": True, "temperature": 0.7, "top_k": 50, "top_p": 0.95,
#                                               "eos_token_id": [2, 64795, 64797]}},
#     "Qwen/Qwen1.5-7B-Chat": {"context_window": 4096, "max_new_tokens": 256,
#                              "generate_kwargs": {"temperature": 0.7, "top_k": 50, "top_p": 0.95}},
#     "Qwen/Qwen1.5-7B-Chat-GPTQ-Int8": {"context_window": 4096, "max_new_tokens": 256,
#                              "generate_kwargs": {"temperature": 0.7, "top_k": 50, "top_p": 0.95}},
#     "baichuan-inc/Baichuan2-7B-Chat": {"context_window": 4096, "max_new_tokens": 256, "generate_kwargs": None},
#     "tiiuae/falcon-7b-instruct": {"context_window": 4096, "max_new_tokens": 256,
#                                   "generate_kwargs": {"do_sample": True, "temperature": 0.7, "top_k": 50,
#                                                       "top_p": 0.95}},
#     "mosaicml/mpt-7b-chat": {"context_window": 4096, "max_new_tokens": 256,
#                              "generate_kwargs": {"do_sample": True, "temperature": 0.7, "top_k": 50, "top_p": 0.95}},
#     "01-ai/Yi-6B-Chat": {"context_window": 4096, "max_new_tokens": 256,
#                          "generate_kwargs": {"temperature": 0.7, "top_k": 50, "top_p": 0.95}},
# }

llm_argument_dict = {
    "meta-llama/Llama-2-7b-chat-hf": {"context_window": 4096, "generate_kwargs": {"temperature": 0}},
    "THUDM/chatglm3-6b": {"context_window": 4096, "generate_kwargs": {"temperature": 0, "eos_token_id": [2, 64795, 64797]}},
    "Qwen/Qwen1.5-7B-Chat": {"context_window": 4096, "generate_kwargs": {"temperature": 0}},
    "Qwen/Qwen1.5-7B-Chat-GPTQ-Int8": {"context_window": 4096, "generate_kwargs": {"temperature": 0}},
    "baichuan-inc/Baichuan2-7B-Chat": {"context_window": 4096, "generate_kwargs": {"temperature": 0}},
    "tiiuae/falcon-7b-instruct": {"context_window": 4096, "generate_kwargs": {"temperature": 0}},
    "mosaicml/mpt-7b-chat": {"context_window": 4096, "generate_kwargs": {"temperature": 0}},
    "01-ai/Yi-6B-Chat": {"context_window": 4096, "generate_kwargs": {"temperature": 0}},
}

def get_huggingfacellm(name):
    print("name is " + name)
    tokenizer, model = tokenizer_and_model_fn_dict[name](name)

    # Create a HF LLM using the llama index wrapper
    llm = HuggingFaceLLM(context_window=llm_argument_dict[name]["context_window"],
                         max_new_tokens=llm_argument_dict[name]["max_new_tokens"],
                         completion_to_prompt=completion_to_prompt_dict[name],
                         generate_kwargs=llm_argument_dict[name]["generate_kwargs"],
                         model=model,
                         tokenizer=tokenizer,
                         device_map="auto", )
    return llm



import logging
import sys

import numpy as np
from deepeval.metrics import (
    ContextualPrecisionMetric,
    ContextualRecallMetric,
    ContextualRelevancyMetric,
    FaithfulnessMetric,
    AnswerRelevancyMetric,
    HallucinationMetric,
    BiasMetric,
    ToxicityMetric,
)
from deepeval.test_case import LLMTestCase
from llama_index.core import PromptTemplate
from llama_index.core.schema import NodeWithScore, TextNode
from uptrain import Settings, Evals, EvalLlamaIndex, operators
from uptrain.framework import DataSchema
from llama_index.core.evaluation import FaithfulnessEvaluator, CorrectnessEvaluator, GuidelineEvaluator
from llama_index.core.evaluation import AnswerRelevancyEvaluator, RelevancyEvaluator, SemanticSimilarityEvaluator

import typing as t
import polars as pl
from uptrain.framework.evals import ParametricEval, ResponseMatching
from uptrain.framework.evalllm import EvalLLM

import nest_asyncio

LLAMA_CUSTOM_FAITHFULNESS_TEMPLATE = PromptTemplate(
    "Please tell if the context supports the given information related to the question.\n"
    "You need to answer with either YES or NO.\n"
    "Answer YES if the context supports the information related to the question, even if most of the context is unrelated. \n"
    "See the examples below.\n"
    "Question: Is apple pie usually double-crusted?\n"
    "Information: Apple pie is generally double-crusted.\n"
    "Context: An apple pie is a fruit pie in which the principal filling ingredient is apples. \n"
    "Apple pie is often served with whipped cream, ice cream ('apple pie à la mode'), custard or cheddar cheese.\n"
    "It is generally double-crusted, with pastry both above and below the filling; the upper crust may be solid or latticed (woven of crosswise strips).\n"
    "Answer: YES\n"
    "Question: Does apple pie usually taste bad?\n"
    "Information: Apple pies tastes bad.\n"
    "Context: An apple pie is a fruit pie in which the principal filling ingredient is apples. \n"
    "Apple pie is often served with whipped cream, ice cream ('apple pie à la mode'), custard or cheddar cheese.\n"
    "It is generally double-crusted, with pastry both above and below the filling; the upper crust may be solid or latticed (woven of crosswise strips).\n"
    "Answer: NO\n"
    "Question and Information: {query_str}\n"
    "Context: {context_str}\n"
    "Answer: "
)


class EvaluationResult_LLM:
    def __init__(self, metrics=None):
        self.results = {}
        # 初始化用于评估的指标，每一个指标都是一个包含score和count的字典
        evaluation_metrics = ["Llama_retrieval_Faithfulness", "Llama_retrieval_Relevancy", "Llama_response_correctness",
                              "Llama_response_semanticSimilarity", "Llama_response_answerRelevancy",
                              "Llama_retrieval_RelevancyG",
                              "Llama_retrieval_FaithfulnessG",
                              "DeepEval_retrieval_contextualPrecision", "DeepEval_retrieval_contextualRecall",
                              "DeepEval_retrieval_contextualRelevancy", "DeepEval_retrieval_faithfulness",
                              "DeepEval_response_answerRelevancy", "DeepEval_response_hallucination",
                              "DeepEval_response_bias", "DeepEval_response_toxicity",
                              "UpTrain_Response_Completeness", "UpTrain_Response_Conciseness",
                              "UpTrain_Response_Relevance",
                              "UpTrain_Response_Valid", "UpTrain_Response_Consistency",
                              "UpTrain_Response_Response_Matching",
                              "UpTrain_Retrieval_Context_Relevance", "UpTrain_Retrieval_Context_Utilization",
                              "UpTrain_Retrieval_Factual_Accuracy", "UpTrain_Retrieval_Context_Conciseness",
                              "UpTrain_Retrieval_Code_Hallucination",
                              ]
        self.evaluationName = evaluation_metrics
        for i in Map_Uptrain_metrics_truth_val.keys():
            evaluation_metrics.append(i)
        evaluation_metrics_rev = []
        for i in evaluation_metrics:
            evaluation_metrics_rev.append(i + "_rev")
        self.metrics_results = {}
        for metric in evaluation_metrics:
            self.metrics_results[metric] = {"score": 0, "count": 0}
            self.metrics_results[metric + "_rev"] = {"score": 0, "count": 0}

        if metrics is None:
            metrics = []

        self.metrics = metrics

    def add(self, evaluate_result):
        for key in self.results.keys():
            if key in self.metrics:
                self.results[key] += evaluate_result.results[key]
        for key in self.metrics_results.keys():
            if key in self.metrics:
                if evaluate_result.metrics_results[key]["score"] != None:
                    self.metrics_results[key]["score"] += evaluate_result.metrics_results[key]["score"]
                    self.metrics_results[key]["count"] += evaluate_result.metrics_results[key]["count"]
        self.results["n"] += 1

    def print_results(self):
        for key, value in self.results.items():
            if key in self.metrics:
                if key == 'n':
                    print(f"{key}: {value}")
                else:
                    print(f"{key}: {value / self.results['n']}")
        for key, value in self.metrics_results.items():
            if key in self.metrics:
                if value['count'] == 0:
                    print(f"{key}: 0, valid number : {value['count']}")
                else:
                    print(f"{key}: {value['score'] / value['count']}, valid number : {value['count']}")

    def print_results_to_path(self, path, config, sample_arr):
        print("save data to " + path)
        f = open(path, 'a')
        f.writelines("\n")
        f.writelines("=========================== begin ==================\n")
        f.writelines("database: " + config.dataset + "\n")
        f.writelines("sample_arr: " + str(sample_arr) + "\n")
        for key, value in self.results.items():
            if key in self.metrics:
                if key == 'n':
                    f.writelines(f"{key}: {value}\n")
                else:
                    f.writelines(f"{key}: {value / self.results['n']}\n")
        for key, value in self.metrics_results.items():
            if key in self.metrics:
                if value['count'] == 0:
                    f.writelines(f"{key}: 0, valid number : {value['count']}\n")
                else:
                    f.writelines(f"{key}: {value['score'] / value['count']}, valid number : {value['count']}\n")


# 在使用uptrain本地模型前，请下载ollama
def upTrain_evaluate_self(
        settings,
        data: t.Union[list[dict], pl.DataFrame],
        checks: list[t.Union[str, Evals, ParametricEval]],
        project_name: str = None,
        schema: t.Union[DataSchema, dict[str, str], None] = None,
        metadata: t.Optional[dict[str, str]] = None,
):
    client = EvalLLM(settings)

    nest_asyncio.apply()

    results = client.evaluate(
        data=data, checks=checks, schema=schema, metadata=metadata
    )
    return results


# 咱们定义的指标对应Uptrain中的指标
Map_Uptrain_metrics_truth_val = {
    "UpTrain_Response_Completeness": Evals.RESPONSE_COMPLETENESS,
    "UpTrain_Response_Conciseness": Evals.RESPONSE_CONCISENESS,
    "UpTrain_Response_Relevance": Evals.RESPONSE_RELEVANCE,
    "UpTrain_Response_Valid": Evals.VALID_RESPONSE,
    "UpTrain_Response_Consistency": Evals.RESPONSE_CONSISTENCY,
    "UpTrain_Response_Response_Matching": ResponseMatching(method='llm'),

    "UpTrain_Retrieval_Context_Relevance": Evals.CONTEXT_RELEVANCE,
    "UpTrain_Retrieval_Context_Utilization": Evals.RESPONSE_COMPLETENESS_WRT_CONTEXT,
    "UpTrain_Retrieval_Factual_Accuracy": Evals.FACTUAL_ACCURACY,
    "UpTrain_Retrieval_Context_Conciseness": Evals.CONTEXT_CONCISENESS,
    "UpTrain_Retrieval_Code_Hallucination": Evals.CODE_HALLUCINATION,

}
# 指标对应得分的名字
Map_Uptrain_metrics_score_name = {
    "UpTrain_Response_Completeness": "score_response_completeness",
    "UpTrain_Response_Conciseness": "score_response_conciseness",
    "UpTrain_Response_Relevance": "score_response_relevance",
    "UpTrain_Response_Valid": "score_valid_response",
    "UpTrain_Response_Consistency": "score_response_consistency",
    "UpTrain_Response_Response_Matching": "score_response_match_recall",

    "UpTrain_Retrieval_Context_Relevance": "score_context_relevance",
    "UpTrain_Retrieval_Context_Utilization": "score_response_completeness_wrt_context",
    "UpTrain_Retrieval_Factual_Accuracy": "score_factual_accuracy",
    "UpTrain_Retrieval_Context_Conciseness": "score_context_conciseness",
    "UpTrain_Retrieval_Code_Hallucination": "score_code_hallucination",
}

#


def UptrainEvaluate(evalModelAgent, question, actual_response, retrieval_context, expected_answer, gold_context, checks,
                    local_model="qwen:7b-chat-v1.5-q8_0"):
    data = list()
    # 这块我把Uptrain里面的评测函数提取出来了，retrieval_context是拼接成字符串作为参数的
    retrieval_context_str = "\n".join(
        [c for c in retrieval_context]
    )
    golden_context_str = "\n".join(
        [c for c in gold_context]
    )
    data.append({"question": question,
                 "response": actual_response,
                 "context": retrieval_context_str,
                 "concise_context": golden_context_str,
                 "ground_truth": expected_answer
                 })
    # 指标所需要的变量会在实现中自己取得
    # settings是使用model的设置， data是指标所需的数据， checks 是需要检测的指标的list
    results = upTrain_evaluate_self(settings=evalModelAgent.uptrainSetting, data=data, checks=checks)
    return results


def get_DeepEval_Metrices(evalModelAgent, model_name="DeepEval_retrieval_contextualPrecision"):
    match model_name:
        case "DeepEval_retrieval_contextualPrecision":
            return ContextualPrecisionMetric(threshold=0.7, model=evalModelAgent.deepEvalModel, include_reason=True)
        case "DeepEval_retrieval_contextualRecall":
            return ContextualRecallMetric(threshold=0.7, model=evalModelAgent.deepEvalModel, include_reason=True)
        case "DeepEval_retrieval_contextualRelevancy":
            return ContextualRelevancyMetric(threshold=0.7, model=evalModelAgent.deepEvalModel, include_reason=True)
        case "DeepEval_retrieval_faithfulness":
            return FaithfulnessMetric(threshold=0.7, model=evalModelAgent.deepEvalModel, include_reason=True)
        case "DeepEval_response_answerRelevancy":
            return AnswerRelevancyMetric(threshold=0.7, model=evalModelAgent.deepEvalModel, include_reason=True)
        case "DeepEval_response_hallucination":
            return HallucinationMetric(threshold=0.7, model=evalModelAgent.deepEvalModel, include_reason=True)
        case "DeepEval_response_bias":
            return BiasMetric(threshold=0.7, model=evalModelAgent.deepEvalModel, include_reason=True)
        case "DeepEval_response_toxicity":
            return ToxicityMetric(threshold=0.7, model=evalModelAgent.deepEvalModel, include_reason=True)


# todo: ours evaluator
# class Evaluator(BaseEvaluator):
def get_llama_evaluator(evalModelAgent, model_name="Llama_retrieval_Faithfulness"):
    match model_name:
        case "Llama_retrieval_Faithfulness":
            return FaithfulnessEvaluator(llm=evalModelAgent.llamaModel)
        case "Llama_retrieval_FaithfulnessG":
            return FaithfulnessEvaluator(llm=evalModelAgent.llamaModel,
                                         eval_template=LLAMA_CUSTOM_FAITHFULNESS_TEMPLATE)
        case "Llama_retrieval_Relevancy":
            return RelevancyEvaluator(llm=evalModelAgent.llamaModel)
        case "Llama_retrieval_RelevancyG":
            return RelevancyEvaluator(llm=evalModelAgent.llamaModel)
        case "Llama_response_correctness":
            return CorrectnessEvaluator(llm=evalModelAgent.llamaModel)
        case "Llama_response_semanticSimilarity":
            return SemanticSimilarityEvaluator()
        case "Llama_response_answerRelevancy":
            return AnswerRelevancyEvaluator(llm=evalModelAgent.llamaModel)


# response evaluate
def evaluating(question, response, actual_response, retrieval_context, retrieval_ids, expected_answer, golden_context,
               golden_context_ids, metrics, evalModelAgent):
    # 创建一个新类，主要是用来记录各个指标有效的个数以及得分
    eval_result = EvaluationResult_LLM()
    # endregion
    # 由于upTrain可以一次计算多个指标，所以这个变量之后会从upTrain的多个指标
    upTrain_metrics = list()
    # region llama_index evaluation
    for i in eval_result.evaluationName:
        if i in metrics and i[0:8] != "DeepEval" and i[0:7] != "UpTrain" and i[0:3] != "NLG":
            print("now run " + i)

            count = 2
            while True:
                try:
                    evaluator = get_llama_evaluator(evalModelAgent, i)
                    res = EvaluationResult_LLM()
                    res.passing = False
                    if i == "Llama_retrieval_FaithfulnessG":
                        # 这块跟prompt相关（LLAMA_CUSTOM_FAITHFULNESS_TEMPLATE）
                        query_str = f"Question: {question}\nInformation: {response.response}"
                        # Faithfulness.evaluate_response的实现中会把query的值传给上面prompt的query_str
                        # 实现的功能和文档中相同，这里试用了一下gpt生成的prompt
                        res = evaluator.evaluate_response(query=query_str, response=response)
                    elif i == "Llama_retrieval_RelevancyG":
                        # 这块要评测golden_context和其他的关系，相当于在原基础上更换retrieval_context
                        # 在Relevancy的实现中，会取出node里面的文本，所以我们构建的时候只传文本就好
                        # 这块语义差不多所以就没更换prompt
                        response.source_nodes.clear()
                        for context in golden_context:
                            node = TextNode()
                            node.text = context
                            temp = NodeWithScore(node=node)
                            response.source_nodes.append(temp)
                        res = evaluator.evaluate_response(query=question, response=response)
                    else:
                        # 这里的传参主要是想要使用系统自带的函数，在实现中会自动提取所需的内容
                        res = evaluator.evaluate_response(query=question, response=response, reference=expected_answer)
                    if res.passing:
                        eval_result.metrics_results[i]["score"] = 1
                    elif 0 <= res.score and res.score <= 1:
                        eval_result.metrics_results[i]["score"] = res.score
                    # 这里要记录一下数量，因为由于种种原因这个过程会报错，就会导致无效的记录
                    eval_result.metrics_results[i]["count"] = 1
                    break
                except Exception as e:
                    count = count - 1

                    logging.exception(e)
                    print("error ")
                    if count == 0:
                        print(i + " error")
                        break
    # endregion
    # region uptrain evaluation
    for i in metrics:
        if i in Map_Uptrain_metrics_truth_val.keys():
            upTrain_metrics.append(i)
    if upTrain_metrics.__len__() != 0:
        # upTrain 可以进行多个指标的评测（传入要评测指标的list）,因此我们将要测的指标封装成一个list
        upTrain_metrics_val = list()
        for i in upTrain_metrics:
            upTrain_metrics_val.append(Map_Uptrain_metrics_truth_val[i])
        count = 2
        while True:
            try:
                result = UptrainEvaluate(evalModelAgent, question, actual_response, retrieval_context, expected_answer,
                                         golden_context, upTrain_metrics_val)
                for i in upTrain_metrics:
                    try:
                        # TODO: fix NONETYPE
                        if result[0][Map_Uptrain_metrics_score_name[i]] >= 0 and result[0][
                            Map_Uptrain_metrics_score_name[i]] <= 1:
                            eval_result.metrics_results[i]["score"] = result[0][Map_Uptrain_metrics_score_name[i]]
                            eval_result.metrics_results[i]["count"] = 1
                    except Exception as e:
                        logging.exception(e)
                break
            except Exception as e:
                count = count - 1
                logging.exception(e)
                if count == 0:
                    break
    # endregion
    # region deepEval evaluation
    for i in eval_result.evaluationName:
        if i in metrics and i[0:8] == "DeepEval":
            count = 2
            while True:
                try:
                    # 构建评测的参数
                    test_case = LLMTestCase(
                        input=question,
                        actual_output=actual_response,
                        retrieval_context=retrieval_context,
                        expected_output=expected_answer,
                        context=golden_context,
                    )
                    deepeval_metric = get_DeepEval_Metrices(evalModelAgent, i)
                    deepeval_metric.measure(test_case)
                    if len(deepeval_metric.verdicts) == 0:
                        raise Exception("deepeval verdicts is zero")

                    if deepeval_metric.score:
                        eval_result.metrics_results[i]["score"] = 1
                    eval_result.metrics_results[i]["count"] = 1
                    break
                except Exception as e:
                    count = count - 1

                    logging.exception(e)
                    #   file2.write(traceback.format_exc())
                    print("error ")
                    if count == 0:
                        print(i + " error")
                        break
    # # region NLG evaluation
    # NLG_metrics = []
    # for i in metrics:
    #     if i[0:3] == "NLG":
    #         NLG_metrics.append(i[4:])
    # if NLG_metrics.__len__() != 0:
    #     result = NLGEvaluate(question, actual_response, golden_context, golden_context_ids, NLG_metrics)
    #     for i in NLG_EVALUATION_METRICS:
    #         eval_result.metrics_results["NLG_" + i]["score"] = result[i]
    #         eval_result.metrics_results["NLG_" + i]["count"] = 1

    # endregion
    return eval_result

# region commonly used indicators



from llama_index.core import VectorStoreIndex
from llama_index.core import (
    StorageContext,
    load_index_from_storage,
)
from llama_index.core.node_parser import SentenceSplitter
from llama_index.core.storage.docstore import SimpleDocumentStore

from ..data.qa_loader import get_documents
import os
from langchain.text_splitter import RecursiveCharacterTextSplitter
from llama_index.core.node_parser import LangchainNodeParser
from llama_index.core.node_parser import HierarchicalNodeParser

def get_index(documents, persist_dir, split_type="sentence", chunk_size=1024):
    hierarchical_storage_context = None
    if not os.path.exists(persist_dir):
        # load the documents and create the index
        if split_type == "sentence":
            parser = SentenceSplitter(chunk_size=chunk_size, chunk_overlap=20)
            nodes = parser.get_nodes_from_documents(documents, show_progress=True)
            print("nodes: " + str(nodes.__len__()))
            index = VectorStoreIndex(nodes,show_progress=True)
        elif split_type == "character":
            parser = LangchainNodeParser(RecursiveCharacterTextSplitter())
            nodes = parser.get_nodes_from_documents(documents, show_progress=True)
            print("nodes: " + str(nodes.__len__()))
            index = VectorStoreIndex(nodes,show_progress=True)
        elif split_type == "hierarchical":
            parser = HierarchicalNodeParser.from_defaults(
                chunk_sizes=[2048, 512, 128]
            )
            nodes = parser.get_nodes_from_documents(documents, show_progress=True)
            print("nodes: " + str(nodes.__len__()))
            index = VectorStoreIndex(nodes,show_progress=True)
        else:
            raise ValueError(f"split_type {split_type} not supported.")
        # store it for later
        if split_type == "hierarchical":
            docstore = SimpleDocumentStore()
            docstore.add_documents(nodes)
            hierarchical_storage_context = StorageContext.from_defaults(docstore=docstore)
            # save
            hierarchical_storage_context.persist(persist_dir=persist_dir+"-hierarchical")

        index.storage_context.persist(persist_dir=persist_dir)
    else:
        # load the existing index
        if split_type == "hierarchical":
            hierarchical_storage_context = StorageContext.from_defaults(persist_dir=persist_dir + "-hierarchical")
        storage_context = StorageContext.from_defaults(persist_dir=persist_dir)
        index = load_index_from_storage(storage_context)
    return index, hierarchical_storage_context



# pip install llama-index-retrievers-bm25

import os
import warnings
from typing import List, Optional

import openai
from llama_index.core.base.base_retriever import BaseRetriever
from llama_index.core.indices.document_summary import DocumentSummaryIndexEmbeddingRetriever, \
    DocumentSummaryIndexLLMRetriever
from llama_index.core.indices.keyword_table.retrievers import BaseKeywordTableRetriever, KeywordTableGPTRetriever
from llama_index.core.indices.list import SummaryIndexEmbeddingRetriever, SummaryIndexRetriever
from llama_index.core.indices.tree import TreeAllLeafRetriever, TreeSelectLeafRetriever, \
    TreeSelectLeafEmbeddingRetriever, TreeRootRetriever
from llama_index.core.postprocessor import MetadataReplacementPostProcessor, SimilarityPostprocessor, LongContextReorder
from llama_index.core.query_engine import RetrieverQueryEngine

import logging
import sys

from llama_index.core import (
    SimpleDirectoryReader,
    VectorStoreIndex, Settings, SummaryIndex, TreeIndex, QueryBundle, StorageContext, DocumentSummaryIndex,
)
from llama_index.core.response_synthesizers import ResponseMode
from llama_index.core.retrievers.fusion_retriever import FUSION_MODES
from llama_index.core.schema import NodeWithScore, IndexNode
from llama_index.core.storage.docstore import SimpleDocumentStore
from llama_index.legacy.indices.keyword_table import KeywordTableSimpleRetriever
from llama_index.retrievers.bm25 import BM25Retriever
from llama_index.core.retrievers import VectorIndexRetriever, QueryFusionRetriever, AutoMergingRetriever, \
    RecursiveRetriever
from llama_index.core.node_parser import SentenceSplitter, SentenceWindowNodeParser
from llama_index.llms.openai import OpenAI

from llama_index.core import get_response_synthesizer

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logging.getLogger().handlers = []
logging.getLogger().addHandler(logging.StreamHandler(stream=sys.stdout))


# todo 常规检索模式: bm25检索 向量检索
def bm25_retriever(index):
    retriever_bm25 = BM25Retriever.from_defaults(index=index, similarity_top_k=3)
    return retriever_bm25


def vector_retriever(index):
    # index = VectorStoreIndex(node)
    retriever_vector = VectorIndexRetriever(index=index, similarity_top_k=3, show_progress=True,
                                            store_nodes_override=True)
    return retriever_vector


# todo 特定于某种索引的检索器类:汇总索引检索 树索引检索 关键字表索引检索 文档摘要索引检索
# note index必须为汇总索引 https://docs.llamaindex.ai/en/stable/api_reference/indices/list.html#llama_index.core.indices.list.SummaryIndex
def summary_retriever(summary_index, mode:int = 0):
    if mode > 2 or mode < 0:
        raise ValueError("Invalid mode for summary retriever."+ str(mode))
    if mode == 0:
        # 最简单的文档汇总索引
        retriever_summary = SummaryIndexRetriever(index=summary_index)
    elif mode == 1:
        # 基于编码的文档汇总索引
        retriever_summary = SummaryIndexEmbeddingRetriever(index=summary_index, embed_model=Settings.embed_model,
                                                           similarity_top_k=3)
    elif mode == 2:
        # 基于大模型的文档汇总索引
        retriever_summary = SummaryIndexEmbeddingRetriever(index=summary_index, llm=Settings.llm, similarity_top_k=3)

    return retriever_summary


# note index必须为树索引 https://docs.llamaindex.ai/en/latest/api_reference/indices/tree.html#llama_index.core.indices.tree.TreeIndex
def tree_retriever(index, mode=0):
    if mode > 3 or mode < 0:
        mode = 0

    # 只使用叶节点的树索引
    if mode == 1:
        retriever_tree = TreeAllLeafRetriever(index=index)
    # 使用部分叶节点
    elif mode == 2:
        retriever_tree = TreeSelectLeafRetriever(index=index)
    # 使用编码结合叶节点的检索器
    elif mode == 3:
        retriever_tree = TreeSelectLeafEmbeddingRetriever(index=index, embed_model=Settings.embed_model)
    # 使用根节点的检索器
    elif mode == 0:
        retriever_tree = TreeRootRetriever(index=index)

    return retriever_tree


# node index必须为关键字表索引 https://docs.llamaindex.ai/en/stable/api_reference/query/retrievers/table.html
def keyword_retriever(index, mode=0):
    if mode > 1 or mode < 0:
        raise ValueError("Invalid mode for keyword retriever."+ str(mode))
    # 基本关键字表检索器

    # GPT关键字表检索器

    retriever_keyword = KeywordTableGPTRetriever(index)
    return retriever_keyword


# 关于该检索器可能的bug：https://github.com/run-llama/llama_index/issues/7633
def document_summary_retrievers(index):
    retriever_d = DocumentSummaryIndexLLMRetriever(
        index,
        choice_batch_size=10,
        choice_top_k=1,
    )
    retriever_d = DocumentSummaryIndexEmbeddingRetriever(index)
    return retriever_d


# todo 高级检索模式：自定义检索器 融合检索 自动合并检索（扩增上下文检索） 元数据替换+句子窗口检索 递归检索
# 自定义检索器，定义了向量检索与bm25检索混合，混合模式为取合
class CustomRetriever(BaseRetriever):
    def __init__(
            self,
            vector_retriever_c: VectorIndexRetriever,
            bm25_retriever_c: BM25Retriever,
            keyword_retriever_c: KeywordTableSimpleRetriever,
            mode: str = "AND",
    ) -> None:
        self._vector_retriever = vector_retriever_c
        self._bm25_retriever = bm25_retriever_c
        self._keyword_retriever = keyword_retriever_c
        if mode not in ("AND", "OR"):
            raise ValueError("Invalid mode.")
        self._mode = mode
        super().__init__()

    def _retrieve(self, query_bundle: QueryBundle) -> List[NodeWithScore]:
        vector_nodes = self._vector_retriever.retrieve(query_bundle)
        bm25_nodes = self._bm25_retriever.retrieve(query_bundle)

        vector_ids = {n.node.node_id for n in vector_nodes}
        bm25_ids = {n.node.node_id for n in bm25_nodes}

        combined_dict = {n.node.node_id: n for n in vector_nodes}
        combined_dict.update({n.node.node_id: n for n in bm25_nodes})

        if self._mode == "AND":
            retrieve_ids = vector_ids.intersection(bm25_ids)
        else:
            retrieve_ids = vector_ids.union(bm25_ids)

        retrieve_nodes = [combined_dict[rid] for rid in retrieve_ids]
        return retrieve_nodes


# 融合检索器 其将来自多个文档的索引作为输入，并自动进行问题扩充，以获得多次查询结果用于合并
# mode: 代表融合模式. 0 代表简单合并. 1: 采用RRF倒数排名融合
def query_fusion_retriever(index, num_queries=4, similarity_top_k=2, mode=0, retriever_weight=None):
    query_fusion_r = None
    if not isinstance(index, list):
        index = [index]
    if retriever_weight is None:
        retriever_weight = [1 / len(index)] * len(index)
    if mode == 0:
        query_fusion_r = QueryFusionRetriever(
            [index_s.as_retriever() for index_s in index],
            llm=Settings.llm,
            similarity_top_k=similarity_top_k,
            num_queries=num_queries,
            use_async=True,
            verbose=True,
            # query_gen_prompt="...",
            # 默认用于多轮询问的问题模板：
            # QUERY_GEN_PROMPT=(
            #     "You are a helpful assistant that generates multiple search queries based on a "
            #     "single input query. Generate {num_queries} search queries, one on each line, "
            #     "related to the following input query:\n"
            #     "Query: {query}\n"
            #     "Queries:\n"
            # )
        )
    elif mode == 1:
        query_fusion_r = QueryFusionRetriever(
            [index_s.as_retriever() for index_s in index],
            llm=Settings.llm,
            similarity_top_k=similarity_top_k,
            num_queries=num_queries,
            use_async=True,
            verbose=True,
            mode=FUSION_MODES.RECIPROCAL_RANK
        )

    return query_fusion_r


# 自动合并检索
# 其将自动合并子节点为高级节点，并进行扩增上下文操作 故需要提供节点类
# 也可使用特殊节点类进行初始化 https://docs.llamaindex.ai/en/latest/examples/retrievers/auto_merging_retriever.html
def auto_merging_retriever(index, hierarchical_storage_context):

    auto_merging_r = AutoMergingRetriever(index.as_retriever(similarity_top_k=6), storage_context=hierarchical_storage_context,
                                          verbose=True)
    return auto_merging_r


# 递归检索 + 句子节点引用rong
def recursive_retriever(base_nodes):
    sub_chunk_sizes = [128, 256, 512]
    sub_node_parsers = [
        SentenceSplitter(chunk_size=c, chunk_overlap=20) for c in sub_chunk_sizes
    ]

    all_nodes = []
    for base_node in base_nodes:
        for n in sub_node_parsers:
            sub_nodes = n.get_nodes_from_documents([base_node])
            sub_inodes = [
                IndexNode.from_text_node(sn, base_node.node_id) for sn in sub_nodes
            ]
            all_nodes.extend(sub_inodes)

        # also add original node to node
        original_node = IndexNode.from_text_node(base_node, base_node.node_id)
        all_nodes.append(original_node)

    all_nodes_dict = {n.node_id: n for n in all_nodes}
    vector_index_chunk = VectorStoreIndex(all_nodes, embed_model=Settings.embed_model)
    vector_retriever_chunk = vector_index_chunk.as_retriever(similarity_top_k=3)
    retriever_chunk = RecursiveRetriever(
        "vector",
        retriever_dict={"vector": vector_retriever_chunk},
        node_dict=all_nodes_dict,
        verbose=True,
    )
    return vector_retriever_chunk


# 句子窗口类需要特殊的节点生成的索引
def sentence_window_retriever(index):
    return index.as_retriever(node_postprocessors=[
        MetadataReplacementPostProcessor(target_metadata_key="window")
    ], )


def custom_retriever(index, mode=0):
    vector_r = vector_retriever(index)
    bm25_r = bm25_retriever(index)
    keyword_r = keyword_retriever(index)
    if mode > 1 or mode < 0:
        raise ValueError("Invalid mode for custom retriever."+ str(mode))

    if mode == 0:
        custom_r = CustomRetriever(vector_retriever_c=vector_r, bm25_retriever_c=bm25_r, mode="AND")
    elif mode == 1:
        custom_r = CustomRetriever(vector_retriever_c=vector_r, keyword_retriever_c=keyword_r, mode="OR")
    return custom_r


# refine：通过按顺序遍历每个检索到的文本块来创建和优化答案。 这将为每个节点/检索到的块进行单独的 LLM 调用。
# compact（默认）：事先压缩（连接）块，从而减少 LLM 调用。填充尽可能多的文本（从检索到的块中串联/打包）可以放入上下文窗口
# tree_summarize：根据需要多次使用提示查询 LLM，以便所有串联的块 被查询，从而产生同样多的答案，这些答案本身在 LLM 调用中以递归方式用作块 依此类推，直到只剩下一个块，因此只有一个最终答案。
# simple_summarize：截断所有文本块以适应单个 LLM 提示。适合快速 摘要目的，但可能会因截断而丢失详细信息。
# accumulate：给定一组文本块和查询，将查询应用于每个文本 块，同时将响应累积到数组中。返回 all 的串联字符串。当您需要对每个文本分别运行相同查询时，非常有用
# compact_accumulate：与 accumulate 相同，但会“压缩”每个类似于 的 LLM 提示符，并对每个文本块运行相同查询。compact
def response_synthesizer(mode=0):
    if mode > 7:
        mode = 1
    choose = [ResponseMode.REFINE, ResponseMode.COMPACT, ResponseMode.COMPACT_ACCUMULATE, ResponseMode.ACCUMULATE,
              ResponseMode.TREE_SUMMARIZE, ResponseMode.SIMPLE_SUMMARIZE, ResponseMode.NO_TEXT, ResponseMode.GENERATION]
    response_s = get_response_synthesizer(response_mode=choose[mode])
    return response_s

# 检索器类型包括：
# BM25
# Vector
# Summary: 汇总检索器 mode = 0 1 2 对应 最简单的文档汇总索引 基于编码的文档汇总索引 基于大模型的文档汇总索引 必须为关键字表索引 *
# Keyword: 关键字表检索器 mode = 0 1 对应 基本关键字表检索器 GPT关键字表检索器 必须为关键字表索引 *
# Custom
# QueryFusion: 融合检索器 其将来自多个文档的索引作为输入，并自动进行问题扩充，以获得多次查询结果用于合并,index需为一个list
# mode = 0, 1 分别使用llama simple重排序方法，1采用RRF
# AutoMerging
# Recursive
# SentenceWindow *
# Tree: index必须为树索引 *
# mode: 1,2,3,0 对应 TreeAllLeafRetriever TreeSelectLeafRetriever TreeSelectLeafEmbeddingRetriever TreeRootRetriever
# mode确定检索器的模式
def get_retriver(type: str, index, mode: int = 0, node = None, hierarchical_storage_context = None):
    if type == "BM25":
        retriever = bm25_retriever(index)
    elif type == "Vector":
        retriever = vector_retriever(index)
    elif type == "Summary":
        retriever = summary_retriever(index, mode=mode)
    elif type == "Tree":
        retriever = tree_retriever(index, mode=mode)
    elif type == "Keyword":
        retriever = keyword_retriever(index, mode=mode)
    elif type == "Custom":
        retriever = custom_retriever(index, mode=mode)
    elif type == "QueryFusion":
        retriever = query_fusion_retriever(index, mode=mode)
    elif type == "AutoMerging":
        retriever = auto_merging_retriever(index, hierarchical_storage_context)
    elif type == "Recursive":
        retriever = recursive_retriever(base_nodes=node)
    elif type == "SentenceWindow":
        retriever = sentence_window_retriever(index)
    else:
        raise Exception("retriever not supported: %s" % mode)

    return retriever


def query_expansion(ret, query_number=4, similarity_top_k=10):
    if ret is None:
        ret = []
        warnings.warn("query_expansion未传入检索器")
    return QueryFusionRetriever(
        ret,
        similarity_top_k=similarity_top_k,
        num_queries=query_number,  # set this to 1 to disable query generation
        use_async=True,
        verbose=True,
        # query_gen_prompt="...",  # we could override the query generation prompt here
    )


class AllRetriever:
    _doc = None,
    _nodes = None,
    _index = None,
    _retriever = []
    _query_number = 4,
    _similar_k_top = 10,
    _syn = 0

    def __init__(self, nodes_, vector_index_, summary_index_, tree_index_, keyword_index_, sentence_index_, mode=0):
        self.bm25_retriever = bm25_retriever(vector_index_)
        self._retriever.append(self.bm25_retriever)
        self.vector_retriever = vector_retriever(vector_index_)
        self._retriever.append(self.vector_retriever)
        self.summary_retriever = summary_retriever(summary_index_)
        self._retriever.append(self.summary_retriever)
        self.tree_retriever = tree_retriever(tree_index_)
        self._retriever.append(self.tree_retriever)
        self.keyword_retriever = keyword_retriever(keyword_index_)
        self._retriever.append(self.keyword_retriever)
        self.doc_s_retriever = document_summary_retrievers(vector_index_)
        self._retriever.append(self.doc_s_retriever)

        self.custom_retriever = custom_retriever(vector_index_)
        self._retriever.append(self.custom_retriever)
        self.query_fusion_retriever = query_fusion_retriever(vector_index_, mode=1)
        self._retriever.append(self.query_fusion_retriever)
        self.auto_merging_retriever = auto_merging_retriever(vector_index_, nodes_)
        self._retriever.append(self.auto_merging_retriever)
        self.recursive_retriever = recursive_retriever(nodes_)
        self._retriever.append(self.recursive_retriever)
        self.router_retriever = get_query_engine_by_router(summary_index_, vector_index_, keyword_index_)
        self._retriever.append(self.router_retriever)
        self.sentence_window_retriever = sentence_window_retriever(sentence_index_)
        self._retriever.append(self.sentence_window_retriever)
        self.response_syn = response_synthesizer(mode)

    def query_expansion(self, retriever, query_number: Optional[int], similarity_number: Optional[int]):
        if query_number is not None:
            self._query_number = query_number
        if similarity_number is not None:
            self._similar_k_top = similarity_number
        return query_expansion([retriever], query_number=self._query_number, similarity_top_k=self._similar_k_top)

    def get_response_mode(self, retriever_, mode=0):
        if mode != 0:
            self._syn = mode
        query_e = RetrieverQueryEngine(
            retriever=retriever_,
            response_synthesizer=response_synthesizer(mode),
            node_postprocessors=[LongContextReorder()]
        )
        return query_e


from llama_index.core.query_engine import RouterQueryEngine
from llama_index.core.selectors import PydanticSingleSelector
from llama_index.core.tools import QueryEngineTool
from llama_index.core import VectorStoreIndex, SummaryIndex

# define query engines
...


def get_query_engine_by_router(summary_index=None, vector_index=None, keyword_index=None):
    summary_tool = None
    vector_tool = None
    keyword_tool = None
    if summary_index is None:
        warnings.warn("Summary_index is None")
    else:
        summary_tool = QueryEngineTool.from_defaults(
            query_engine=summary_retriever(summary_index),
            description="Useful for summarization questions related to the data source",
        )
    if vector_index is None:
        warnings.warn("vector_index is None")
    else:
        vector_tool = QueryEngineTool.from_defaults(
            query_engine=custom_retriever(vector_index),
            description="Useful for retrieving specific context related to the data source",
        )

    if keyword_index is None:
        warnings.warn("keyword_index is None")
    else:
        keyword_tool = QueryEngineTool.from_defaults(
            query_engine=keyword_retriever(keyword_index),
            description="Useful for retrieving keyword related to the data source",
        )
    query_engine_ = RouterQueryEngine(
        selector=PydanticSingleSelector.from_defaults(),
        query_engine_tools=[
            summary_tool,
            vector_tool,
            keyword_tool
        ],
    )
    return query_engine_


if __name__ == '__main__':
    Settings.llm = OpenAI(temperature=0.2, model="gpt-3.5-turbo")
    # 需要一个直接放文件的本地目录
    documents = SimpleDirectoryReader(f"E:\\junior_second_\\benchmark_llm\\RAG-benchmark").load_data()
    index_ = VectorStoreIndex.from_documents(documents,
                                    transformations=[SentenceSplitter(chunk_size=512, chunk_overlap=20)],
                                    show_progress=True)
    # splitter = SentenceSplitter(
    #     chunk_size=1024,
    #     chunk_overlap=20,
    # )
    # nodes = splitter.get_nodes_from_documents(documents)
    # index_ = VectorStoreIndex(nodes)
    retriever = vector_retriever(index_)  # 可用
    # retriever = vector_retriever(index_)  # 可用

    # retriever = summary_retriever(SummaryIndex(nodes))  # 可用
    # retriever = summary_retriever(index_)

    # retriever = tree_retriever(TreeIndex(nodes))  # 可用，看起来效果很差

    # retriever = keyword_retriever(index_)  # 未测试

    # vector_retriever = vector_retriever(index=index_)
    # bm25_retriever = bm25_retriever(index=index_)
    # retriever = CustomRetriever(vector_retriever=vector_retriever, bm25_retriever=bm25_retriever)  # 可用

    # retriever = query_fusion_retriever([index_, index_], mode=1)  #可用

    # retriever = auto_merging_retriever(index_, nodes)  # 可用

    #     node_parser = SentenceWindowNodeParser.from_defaults(
    #         window_size=3,
    #         window_metadata_key="window",
    #         original_text_metadata_key="original_text",
    #     )
    # retriever = sentence_window_retriever(index_)  # 可用

    # retriever = recursive_retriever(nodes)  # 可用
    response_syn = response_synthesizer(0)
    nodes = retriever.retrieve("请用中文回答我的毕业设计题目是什么")
    print(nodes)
    query_engine = RetrieverQueryEngine(
        retriever=get_retriver("QueryFusion", index_, mode=0),
        response_synthesizer=response_syn,
        node_postprocessors=[LongContextReorder()]
    )
    query_e = query_expansion([query_engine],query_number=4,similarity_top_k=3)
    query_engine = RetrieverQueryEngine.from_args(query_e)
    response = query_engine.query("请用中文回答我的毕业设计题目是什么")
    print(response)



from llama_index.embeddings.huggingface import HuggingFaceEmbedding
# from llama_index.legacy.embeddings import HuggingFaceEmbedding

def get_embedding(name):
    return HuggingFaceEmbedding(
        model_name=name,
        embed_batch_size=16,
        # cache_folder="./embedding_model"
    )

'''
from langchain.embeddings.huggingface import HuggingFaceEmbeddings

def get_embedding(name):
    encode_kwargs = {"batch_size": 128, 'device': 'cuda'}
    embeddings = HuggingFaceEmbeddings(
        model_name=name,
        encode_kwargs=encode_kwargs,
        # embed_batch_size=128,
    )
    return embeddings
'''


import argparse
import sys
from enum import Enum, unique
from .webui import run_web_ui
from .launcher import run
from .config import Config
from .data.qa_loader import generate_qa_from_folder
VERSION = "0.1.3"
USAGE = (
    "-" * 70
    + "\n"
    + "| Usage:                                                             |\n"
    + "|   xrag-cli run -h: launch an eval experiment       |\n"
    + "|   xrag-cli webui: launch XRAGBoard                        |\n"
    + "|   xrag-cli version: show version info                      |\n"
    + "|   xrag-cli generate -i <input_file> -o <output_file> -n <num_questions> -s <sentence_length>: generate QA pairs from a folder |\n"
    + "|   xrag-cli api --host <host> --port <port> --json_path <json_path> --dataset_folder <dataset_folder>: run the API server |\n"
    + "-" * 70
)

WELCOME = (
    "-" * 58
    + "\n"
    + "| Welcome to XRAG, version {}".format(VERSION)
    + " " * (21 - len(VERSION))
    + "|\n|"
    + " " * 56
    + "|\n"
    + "| Project page: https://github.com/DocAILab/xrag |\n"
    + "-" * 58
)

@unique
class Command(str, Enum):
    RUN = "run"
    WEBUI = "webui"
    VER = "version"
    GENERATE = "generate"
    HELP = "help"
    API = "api"

def main():
    # Initialize the argument parser
    parser = argparse.ArgumentParser(description='XRAG CLI Tool')

    # Define the subcommands
    subparsers = parser.add_subparsers(dest='command', help='Subcommands')

    # 'run' command
    run_parser = subparsers.add_parser('run', help='Run the application')
    run_parser.add_argument('--override', nargs='*', help='Override config values (e.g., --override key1=value1 key2=value2)')
    run_parser.add_argument('-c', '--custom_dataset', default='', type=str, help='Custom dataset json path')

    # Other commands
    subparsers.add_parser('webui', help='Run the web UI')
    subparsers.add_parser('version', help='Show version')
    subparsers.add_parser('help', help='Show help')
    
    # API command
    api_parser = subparsers.add_parser('api', help='Run the API server')
    api_parser.add_argument('--host', type=str, default='0.0.0.0', help='API server host')
    api_parser.add_argument('--port', type=int, default=8000, help='API server port')
    api_parser.add_argument('--json_path', type=str, default='', help='JSON file path')
    api_parser.add_argument('--dataset_folder', type=str, default='', help='Dataset folder path')

    generate_parser = subparsers.add_parser('generate', help='Generate QA pairs')
    generate_parser.add_argument('-i', '--input', type=str, help='Input file path')
    generate_parser.add_argument('-o', '--output', type=str, help='Output file path')
    generate_parser.add_argument('-n', '--num', type=int, help='Number of questions per file', default=3)
    generate_parser.add_argument('-s', '--sentence_length', type=int, help='Sentence length, -1 means no split', default=-1)
    # Parse the arguments
    args = parser.parse_args()

    # Handle commands
    if args.command == Command.RUN:
        # Parse overrides
        config_overrides = {}
        if args.override:
            for override in args.override:
                if '=' in override:
                    key, value = override.split('=', 1)
                    config_overrides[key.strip()] = value.strip()
                else:
                    print(f"Invalid override format: {override}")
                    sys.exit(1)
        # Update the Config instance
        config = Config()
        config.update_config(config_overrides)
        if args.custom_dataset:
            run(custom_dataset=args.custom_dataset)
        else:
            run()
    elif args.command == Command.WEBUI:
        run_web_ui()
    elif args.command == Command.VER:
        print(WELCOME)
    elif args.command == Command.HELP or args.command is None:
        parser.print_help()
    elif args.command == Command.GENERATE:
        generate_qa_from_folder(args.input, args.output, args.num, args.sentence_length)
    elif args.command == Command.API:
        from .api.server import run_api_server
        run_api_server(host=args.host, port=args.port, json_path=args.json_path, dataset_folder=args.dataset_folder)
    else:
        print(f"Unknown command: {args.command}")



import os
import toml


class Config:
    _instance = None  # Singleton instance

    def __new__(cls, config_file_path=None):
        if cls._instance is None:
            cls._instance = super(Config, cls).__new__(cls)
            if config_file_path is None:
                config_file_path = 'config.toml'
            cls._instance.config = toml.load(config_file_path)  # Load the config only once

            # Dynamically set attributes based on TOML config
            for section, values in cls._instance.config.items():
                for key, value in values.items():
                    setattr(cls._instance, key, value)

        return cls._instance

    def update_config(self, overrides):
        for key, value in overrides.items():
            if hasattr(self, key):
                old_value = getattr(self, key)
                new_value = self._convert_type(value, type(old_value))
                setattr(self, key, new_value)
            else:
                print(f"Invalid configuration key: {key}")

    def _convert_type(self, value, to_type):
        try:
            if to_type == bool:
                return value.lower() in ('true', '1', 'yes')
            elif to_type == int:
                return int(value)
            elif to_type == float:
                return float(value)
            elif to_type == str:
                return value
            else:
                return value  # For other types
        except ValueError:
            print(f"Could not convert value '{value}' to type {to_type.__name__}")
            return value


class GlobalVar:
    query_number = 0

    @staticmethod
    def set_query_number(num):
        GlobalVar.query_number = num

    @staticmethod
    def get_query_number():
        return GlobalVar.query_number


if __name__ == '__main__':
    # Usage: Create a global instance
    cfg = Config()

    # Now you can access config values directly as attributes:
    print(cfg.test_init_total_number_documents)  # Outputs: 20
    print(cfg.api_key)  # Outputs the API key from the toml file



from typing import Optional, List, Mapping, Any, Sequence, Dict
from llama_index.core.bridge.pydantic import Field, PrivateAttr
from llama_index.core.constants import DEFAULT_CONTEXT_WINDOW, DEFAULT_NUM_OUTPUTS
from llama_index.core.llms import (
    CustomLLM,
    CompletionResponse,
    CompletionResponseGen,
    LLMMetadata,
    ChatMessage,
    ChatResponse,
)
from llama_index.core.llms.callbacks import llm_completion_callback, llm_chat_callback

from zhipuai import ZhipuAI
# pip install zhipuai
DEFAULT_MODEL = 'glm-4'

def to_message_dicts(messages: Sequence[ChatMessage])->List:
    return [
        {"role": message.role.value, "content": message.content,}
                for message in messages if all([value is not None for value in message.values()])
    ]

def get_additional_kwargs(response) -> Dict:
    return {
        "token_counts":response.usage.total_tokens,
        "prompt_tokens": response.usage.prompt_tokens,
        "completion_tokens": response.usage.completion_tokens,
    }

class ChatGLM(CustomLLM):
    num_output: int = DEFAULT_NUM_OUTPUTS
    context_window: int = Field(default=DEFAULT_CONTEXT_WINDOW,description="The maximum number of context tokens for the model.",gt=0,)
    model: str = Field(default=DEFAULT_MODEL, description="The ChatGlM model to use. glm-4 or glm-3-turbo")
    api_key: str = Field(default=None, description="The ChatGLM API key.")
    reuse_client: bool = Field(default=True, description=(
            "Reuse the client between requests. When doing anything with large "
            "volumes of async API calls, setting this to false can improve stability."
        ),
    )

    _client: Optional[Any] = PrivateAttr()
    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        reuse_client: bool = True,
        api_key: Optional[str] = None,
        **kwargs: Any,
    )-> None:
        super().__init__(
            model=model,
            api_key=api_key,
            reuse_client=reuse_client,
            **kwargs,
        )
        self._client = None

    def _get_client(self) -> ZhipuAI:
        if not self.reuse_client :
            return ZhipuAI(api_key=self.api_key)

        if self._client is None:
            self._client = ZhipuAI(api_key=self.api_key)
        return self._client

    @classmethod
    def class_name(cls) -> str:
        return "chatglm_llm"

    @property
    def metadata(self) -> LLMMetadata:
        """Get LLM metadata."""
        return LLMMetadata(
            context_window=self.context_window,
            num_output=self.num_output,
            model_name=self.model,
        )

    def _chat(self, messages:List, stream=False) -> Any:
        #print("--------------------------------------------")
        #import traceback
        #s=traceback.extract_stack()
        # print("%s %s invoke _chat" % (s[-2],s[-2][2]))
        # print(messages)
        # print("--------------------------------------------")
        response = self._get_client().chat.completions.create(
            model=self.model,  # 填写需要调用的模型名称
            messages=messages,
        )
        # print(f"_chat, response: {response}")
        return response

    #@llm_chat_callback()
    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        message_dicts: List = to_message_dicts(messages)
        response = self._chat(message_dicts, stream=False)
        # print(f"chat: {response} ")
        rsp = ChatResponse(
            message=ChatMessage(content=response.choices[0].message.content, role=MessageRole(response.choices[0].message.role),
                additional_kwargs= {}),
            raw=response, additional_kwargs= get_additional_kwargs(response),
        )
        print(f"chat: {rsp} ")

        return rsp

    #@llm_chat_callback()
    def stream_chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> CompletionResponseGen:
        response_txt = ""
        message_dicts: List = to_message_dicts(messages)
        response = self._chat(message_dicts, stream=True)
        # print(f"stream_chat: {response} ")
        for chunk in response:
            # chunk.choices[0].delta # content='```' role='assistant' tool_calls=None
            token = chunk.choices[0].delta.content
            response_txt += token
            yield ChatResponse(message=ChatMessage(content=response_txt,role=MessageRole(message.get("role")),
                                additional_kwargs={},), delta=token, raw=chunk,)

    @llm_completion_callback()
    def complete(self, prompt: str, **kwargs: Any) -> CompletionResponse:
        messages = [{"role": "user", "content": prompt}]
        # print(f"complete: messages {messages} ")
        try:
            response = self._chat(messages, stream=False)

            rsp = CompletionResponse(text=str(response.choices[0].message.content),
                                     raw=response,
                                     additional_kwargs=get_additional_kwargs(response),)
            # print(f"complete: {rsp} ")
        except Exception as e:
            print(f"complete: exception {e}")

        return rsp

    @llm_completion_callback()
    def stream_complete(self, prompt: str, **kwargs: Any) -> CompletionResponseGen:
        response_txt = ""
        messages = [{"role": "user", "content": prompt}]
        response = self._chat(messages, stream=True)
        # print(f"stream_complete: {response} ")
        for chunk in response:
            # chunk.choices[0].delta # content='```' role='assistant' tool_calls=None
            token = chunk.choices[0].delta.content
            response_txt += token
            yield CompletionResponse(text=response_txt, delta=token)


from typing import Optional, List, Mapping, Any, Sequence, Dict
from llama_index.core.embeddings import BaseEmbedding
from llama_index.core.bridge.pydantic import Field, PrivateAttr
from zhipuai import ZhipuAI

class ChatGLMEmbeddings(BaseEmbedding):
    model: str = Field(default='embedding-2', description="The ChatGlM model to use. embedding-2")
    api_key: str = Field(default=None, description="The ChatGLM API key.")
    reuse_client: bool = Field(default=True, description=(
            "Reuse the client between requests. When doing anything with large "
            "volumes of async API calls, setting this to false can improve stability."
        ),
    )

    _client: Optional[Any] = PrivateAttr()
    def __init__(
        self,
        model: str = 'embedding-2',
        reuse_client: bool = True,
        api_key: Optional[str] = None,
        **kwargs: Any,
    )-> None:
        super().__init__(
            model=model,
            api_key=api_key,
            reuse_client=reuse_client,
            **kwargs,
        )
        self._client = None

    def _get_client(self) -> ZhipuAI:
        if not self.reuse_client :
            return ZhipuAI(api_key=self.api_key)

        if self._client is None:
            self._client = ZhipuAI(api_key=self.api_key)
        return self._client

    @classmethod
    def class_name(cls) -> str:
        return "ChatGLMEmbedding"

    def _get_query_embedding(self, query: str) -> List[float]:
        """Get query embedding."""
        return self.get_general_text_embedding(query)

    async def _aget_query_embedding(self, query: str) -> List[float]:
        """The asynchronous version of _get_query_embedding."""
        return self.get_general_text_embedding(query)

    def _get_text_embedding(self, text: str) -> List[float]:
        """Get text embedding."""
        return self.get_general_text_embedding(text)

    async def _aget_text_embedding(self, text: str) -> List[float]:
        """Asynchronously get text embedding."""
        return self.get_general_text_embedding(text)

    def _get_text_embeddings(self, texts: List[str]) -> List[List[float]]:
        """Get text embeddings."""
        embeddings_list: List[List[float]] = []
        for text in texts:
            embeddings = self.get_general_text_embedding(text)
            embeddings_list.append(embeddings)

        return embeddings_list

    async def _aget_text_embeddings(self, texts: List[str]) -> List[List[float]]:
        """Asynchronously get text embeddings."""
        return self._get_text_embeddings(texts)

    def get_general_text_embedding(self, prompt: str) -> List[float]:
        response = self._get_client().embeddings.create(
            model=self.model, #填写需要调用的模型名称
            input=prompt,
        )
        return response.data[0].embedding



import warnings
from streamlit_card import card
import streamlit as st
import pandas as pd
import plotly.express as px
from xrag.config import Config
from xrag.eval.EvalModelAgent import EvalModelAgent
from xrag.eval.evaluate_rag import evaluating
from xrag.launcher import run
from xrag.eval.evaluate_rag import EvaluationResult
from xrag.process.query_transform import transform_and_query
from xrag.launcher import build_index, build_query_engine
from xrag.data.qa_loader import get_qa_dataset

AVAILABLE_METRICS = [
    "NLG_chrf", "NLG_meteor", "NLG_wer", "NLG_cer", "NLG_chrf_pp",
    "NLG_perplexity",
    "NLG_rouge_rouge1", "NLG_rouge_rouge2", "NLG_rouge_rougeL", "NLG_rouge_rougeLsum",
    "Llama_retrieval_Faithfulness", "Llama_retrieval_Relevancy", "Llama_response_correctness",
    "Llama_response_semanticSimilarity", "Llama_response_answerRelevancy", "Llama_retrieval_RelevancyG",
    "Llama_retrieval_FaithfulnessG",
    "DeepEval_retrieval_contextualPrecision", "DeepEval_retrieval_contextualRecall",
    "DeepEval_retrieval_contextualRelevancy", "DeepEval_retrieval_faithfulness",
    "DeepEval_response_answerRelevancy", "DeepEval_response_hallucination",
    "DeepEval_response_bias", "DeepEval_response_toxicity",
    "UpTrain_Response_Completeness", "UpTrain_Response_Conciseness", "UpTrain_Response_Relevance",
    "UpTrain_Response_Valid", "UpTrain_Response_Consistency", "UpTrain_Response_Response_Matching",
    "UpTrain_Retrieval_Context_Relevance", "UpTrain_Retrieval_Context_Utilization",
    "UpTrain_Retrieval_Factual_Accuracy", "UpTrain_Retrieval_Context_Conciseness",
    "UpTrain_Retrieval_Code_Hallucination",
]

# Define options for each dropdown
LLM_OPTIONS = [
    # huggingface models
    # 'llama',      # meta-llama/Llama-2-7b-chat-hf
    # 'chatglm',    # THUDM/chatglm3-6b
    # 'qwen',       # Qwen/Qwen1.5-7B-Chat
    # 'qwen14_int8',# Qwen/Qwen1.5-14B-Chat-GPTQ-Int8
    # 'qwen7_int8', # Qwen/Qwen1.5-7B-Chat-GPTQ-Int8
    # 'qwen1.8',    # Qwen/Qwen1.5-1.8B-Chat
    # 'baichuan',   # baichuan-inc/Baichuan2-7B-Chat
    # 'falcon',     # tiiuae/falcon-7b-instruct
    # 'mpt',        # mosaicml/mpt-7b-chat
    # 'yi',         # 01-ai/Yi-6B-Chat

    'openai',     # OpenAI API
    'huggingface',# HuggingFace local models
    'ollama',     # Ollama local models
]

HF_MODEL_OPTIONS = [
    'llama',      # meta-llama/Llama-2-7b-chat-hf
    'chatglm',    # THUDM/chatglm3-6b
    'qwen',       # Qwen/Qwen1.5-7B-Chat
    'qwen14_int8',# Qwen/Qwen1.5-14B-Chat-GPTQ-Int8
    'qwen7_int8', # Qwen/Qwen1.5-7B-Chat-GPTQ-Int8
    'qwen1.8',    # Qwen/Qwen1.5-1.8B-Chat
    'baichuan',   # baichuan-inc/Baichuan2-7B-Chat
    'falcon',     # tiiuae/falcon-7b-instruct
    'mpt',        # mosaicml/mpt-7b-chat
    'yi',         # 01-ai/Yi-6B-Chat
]

# ollama cascade dict
OLLAMA_OPTIONS = {
    "LLaMA": {
        "llama2-7b": "llama2:7b",
        "llama2-13b": "llama2:13b",
        "llama2-70b": "llama2:70b",

        "llama3.1-8b": "llama3.1:8b",
        "llama3.1-70b": "llama3.1:70b",

        "llama3.2-1b": "llama3.2:1b",
        "llama3.2-3b": "llama3.2:3b",

    },
    "Mistral": {
        "mistral-7b": "mistral:7b",
        "mixtral": "mixtral",
        "mistral-7b-q4": "mistral:7b-q4_0",
    },
    "others": {
        "gemma-7b": "gemma:7b",
        "codellama": "codellama",
        "neural-chat": "neural-chat",
        "other": "other"
    },
}

EMBEDDING_OPTIONS = ["BAAI/bge-large-en-v1.5", "BAAI/bge-m3", "BAAI/bge-base-en-v1.5","BAAI/bge-small-en-v1.5","BAAI/bge-large-zh-v1.5","BAAI/bge-base-zh-v1.5","BAAI/bge-small-zh-v1.5"]  # Add more as needed
SPLIT_TYPE_OPTIONS = ["sentence", "character", "hierarchical"]
DATASET_OPTIONS = ["hotpot_qa", "drop", "natural_questions","trivia_qa","search_qa","finqa","law"]  # Replace with actual dataset options
DATASET_DISPLAY_MAP = {
    "hotpot_qa": "HotpotQA",
    "drop": "DropQA",
    "natural_questions": "NaturalQA"
    # remain for other options
}

DATASETS_INFO = {
    "HotpotQA": {
        "size": {
            "train": "86,830",
            "validation": "8,680",
            "test": "968"
        },
        "corpus": {
            "documents": "508,826",
            "source": "Wikipedia"
        },
        "features": {
            "Multi-hop": True,
            "Constrained": False,
            "Numerical": True,
            "Set-logical": False
        }
    },
    "DropQA": {
        "size": {
            "train": "78,241",
            "validation": "7,824",
            "test": "870"
        },
        "corpus": {
            "documents": "6,147",
            "source": "Wikipedia"
        },
        "features": {
            "Multi-hop": True,
            "Constrained": False,
            "Numerical": True,
            "Set-logical": True
        }
    },
    "NaturalQA": {
        "size": {
            "train": "100,093",
            "validation": "10,010",
            "test": "1,112"
        },
        "corpus": {
            "documents": "49,815",
            "source": "Wikipedia"
        },
        "features": {
            "Multi-hop": True,
            "Constrained": True,
            "Numerical": True,
            "Set-logical": False
        }
    }
}

# frontend dataset options
FRONTEND_DATASET_OPTIONS = [DATASET_DISPLAY_MAP.get(ds, ds) for ds in DATASET_OPTIONS]
RETRIEVER_OPTIONS = ["BM25", "Vector", "Summary", "Tree", "Keyword", "Custom", "QueryFusion", "AutoMerging", "Recursive", "SentenceWindow"]  # Add more as needed
POSTPROCESS_RERANK_OPTIONS = ["none","long_context_reorder", "colbertv2_rerank","bge-reranker-base"]  # Add more as needed
QUERY_TRANSFORM_OPTIONS = ["none", "hyde_zeroshot", "hyde_fewshot","stepback_zeroshot","stepback_fewshot"]  # Add more as needed

# follow the order listed in the docs
METRIC_DISPLAY_MAP = {
    "NLG_chrf": "ChrF",
    "NLG_chrf_pp": "ChrF++",
    "NLG_meteor": "METEOR",
    "NLG_rouge_rouge1": "ROUGE1",
    "NLG_rouge_rouge2": "ROUGE2",
    "NLG_rouge_rougeL": "ROUGEL",
    "NLG_rouge_rougeLsum": "ROUGELSUM",
    "nlg-em": "EM",
    "NLG_perplexity": "PPL",
    "NLG_cer": "CER",
    "NLG_wer": "WER",
    
    # Llama metrics
    "Llama_retrieval_Faithfulness": "Llama-Response-Faithfulness",
    "Llama_retrieval_Relevancy": "Llama-Response-Relevance",
    "Llama_response_correctness": "Llama-Response-Correctness",
    "Llama_response_semanticSimilarity": "Llama-Response-Similarity",
    "Llama_response_answerRelevancy": "Llama-Response-Relevance++",
    "Llama_retrieval_FaithfulnessG": "Llama-Response-Faithfulness+",
    "Llama_retrieval_RelevancyG": "Llama-Response-Relevance+",
    
    # UpTrain metrics
    "UpTrain_Retrieval_Context_Relevance": "Uptrain-Context-Relevance",
    "UpTrain_Retrieval_Context_Conciseness": "Uptrain-Context-Conciseness",
    "DeepEval_response_answerRelevancy": "DeepEval-Response-Relevancy",
    "UpTrain_Response_Completeness": "Uptrain-Response-Completeness",
    "UpTrain_Response_Conciseness": "Uptrain-Response-Conciseness",
    "UpTrain_Response_Relevance": "Uptrain-Response-Relevance",
    "UpTrain_Response_Valid": "Uptrain-Response-Valid",
    "UpTrain_Response_Response_Matching": "Uptrain-Response-Matching",

    "UpTrain_Retrieval_Code_Hallucination": "Uptrain-Retrieval-Code-Hallucination",
    
    # DeepEval metrics
    "DeepEval_retrieval_contextualPrecision": "DeepEval-Context-Recall",
    "DeepEval_retrieval_contextualRecall": "DeepEval-Context-Relevance",
    "DeepEval_retrieval_contextualRelevancy": "Uptrain-Context-Consistency",
    "UpTrain_Response_Consistency": "Uptrain-Context-Utilization",
    "UpTrain_Retrieval_Context_Utilization": "Uptrain-Factual-Accuracy",
    "UpTrain_Retrieval_Factual_Accuracy": "Uptrain-Factual-Accuracy",
    "DeepEval_retrieval_faithfulness": "DeepEval-Context-Faithfulness",
    "DeepEval_response_hallucination": "DeepEval-Context-Hallucination",
    
    # others not in the docs but in the code
    "DeepEval_response_bias": "DeepEval-Response-Bias",
    "DeepEval_response_toxicity": "DeepEval-Response-Toxicity",
}

# frontend values
FRONTEND_AVAILABLE_METRICS = [METRIC_DISPLAY_MAP.get(metric, metric) for metric in AVAILABLE_METRICS]
# reverse map
DISPLAY_TO_BACKEND_METRIC_MAP = {v: k for k, v in METRIC_DISPLAY_MAP.items()}

@st.cache_resource(show_spinner=False)
def get_query():
    return run(cli=False)

@st.cache_resource(show_spinner=False)
def get_qa_dataset_(dataset,files=None):
    return get_qa_dataset(dataset,files=files)

@st.cache_resource(show_spinner=False)
def get_index():
    return build_index(st.session_state.qa_dataset['documents'])

@st.cache_resource(show_spinner=False)
def get_query_engine():
    return build_query_engine(st.session_state.index, st.session_state.hierarchical_storage_context)


def main():

    if "step" not in st.session_state:
        st.session_state.step = 1
    st.set_page_config(layout="wide")
    st.title("XRAG")
    cfg = Config()

    if st.session_state.step == 1:
        st.header("Choose your Dataset")
        
        # find the backend dataset value
        backend_dataset = cfg.dataset if cfg.dataset in DATASET_OPTIONS else DATASET_OPTIONS[0]
        # find the frontend dataset value and index
        frontend_dataset = DATASET_DISPLAY_MAP.get(backend_dataset, backend_dataset)
        frontend_index = FRONTEND_DATASET_OPTIONS.index(frontend_dataset)

        # selected_frontend_dataset = st.selectbox(
        #     "Dataset",
        #     options=FRONTEND_DATASET_OPTIONS,
        #     index=frontend_index,
        #     key="dataset"
        # )
        if 'dataset' not in st.session_state:
            st.session_state.dataset = "HotpotQA"

        cols = st.columns(3)
        st.markdown("""
        <style>
            .stCard {
                background-color: white !important;
            }
            .card-body {
                background-color: white !important;
            }
            .card {
                background-color: white !important;
            }
            /* 修改 iframe 的高度 */
            iframe.stCustomComponentV1 {
                height: 400px !important;  /* 调整这个值来改变卡片高度 */
            }
            /* 减少卡片之间的间距 */
            div[data-testid="column"] {
                margin-top: 0 !important;
                margin-bottom: 0 !important;
            }
            .block-container {
                padding-top: 2rem !important;
                padding-bottom: 2rem !important;
            }
        </style>
        """, unsafe_allow_html=True)

        for i, (dataset_name, info) in enumerate(DATASETS_INFO.items()):
            with cols[i]:
                is_selected = st.session_state.dataset == dataset_name

                features_text = "\n".join([
                    f"• {k}: {'✓' if v else '✗'}"
                    for k, v in info['features'].items()
                ])

                size_text = (
                    f"Train: {info['size']['train']}\n"
                    f"Val: {info['size']['validation']}\n"
                    f"Test: {info['size']['test']}"
                )

                clicked = card(
                    title=dataset_name,
                    text=(
                        f"📊 Dataset Size:\n{size_text}\n"
                        f"📚 Corpus:\n"
                        f"• Documents: {info['corpus']['documents']}"
                        # f"✨ Features:\n{features_text}"
                    ),
                    image="https://hotpotqa.github.io/img/home-bg.jpg",
                    styles={
                        "card": {
                            "width": "100%",
                            "height": "300px",
                            "border-radius": "10px",
                            "box-shadow": "0 0 10px rgba(0,0,0,0.1)",
                            "background-color": "#ffffff",  # 设置背景色为白色
                            "border": "4px solid #2E86C1" if is_selected else "1px solid #e0e0e0",
                        },
                        "title": {
                            "font-size": "20px",
                            "font-weight": "bold",
                            "color": "#ffffff",  # 标题改为黑色
                            "margin-bottom": "10px"
                        },
                        "text": {
                            "font-size": "14px",
                            "line-height": "1.5",
                            "white-space": "pre-line",
                            "color": "#ffffff"  # 文本改为黑色
                        }
                    }
                )

                if clicked:
                    st.session_state.dataset = dataset_name
                    st.rerun()

        selected_frontend_dataset = st.session_state.dataset

        # reverse map: from the frontend dataset to the backend dataset
        DISPLAY_TO_BACKEND_MAP = {v: k for k, v in DATASET_DISPLAY_MAP.items()}
        chosen_backend_dataset = DISPLAY_TO_BACKEND_MAP.get(selected_frontend_dataset, selected_frontend_dataset)
        cfg.dataset = chosen_backend_dataset
        # custom dataset
        st.markdown("---")
        st.markdown("## Or your own dataset")
        tab1, tab2 = st.tabs(["Upload JSON", "Generate from Folder"])
        
        with tab1:
            with st.expander("About Custom Dataset"):
                st.markdown("""
                ### Upload Custom Dataset
                
                Please upload a JSON file containing your dataset. The JSON file should be a list of objects with the following format:
                ```json
                [
                    {
                        "question": "What is the capital of France?",
                        "answer": "Paris",
                        "file_paths": "path/to/document.txt",
                        "source_text": "source of the document"
                        // or multiple files
                        // "file_paths": ["path/to/doc1.txt", "path/to/doc2.txt"]
                        // "source": ["source of the document 1", "source of the document 2"]
                    },
                    ...
                ]
                ```
                Note: 
                1. Make sure all file paths in the JSON are accessible from the server.
                2. Supported file formats: txt, md, pdf, html, json, csv, etc.
                3. Each question can reference one or multiple documents.
                4. The system will automatically process and index all documents.
                """)
            
            uploaded_file = st.file_uploader("Upload your dataset", type=["json"])
        
        with tab2:
            st.markdown("""
            ### Generate QA Pairs from Documents
            
            Upload a folder containing your documents, and the system will:
            1. Read all documents in the folder (including subfolders)
            2. Use AI to generate relevant questions and answers
            3. Create a dataset in the required format
            
            Supported file formats:
            - Text files (.txt)
            - Markdown files (.md)
            - PDF documents (.pdf)
            - HTML files (.html)
            - And more...
            
            Note:
            - If the sentence length is set to -1, the system will use file level as the unit.
            - If the sentence length is set to a positive number, the system will split the document into chunks of the specified length.
            """)
            
            folder_path = st.text_input("Enter folder path", value="./data/documents")
            num_questions = st.number_input("Number of questions per file", min_value=1, value=3)
            output_file = st.text_input("Output JSON file path", value="./data/generated_qa.json")
            sentence_length = st.number_input("Sentence length", value=-1)
            if st.button("Generate QA Dataset"):
                try:
                    with st.spinner("Generating QA pairs from documents..."):
                        from xrag.data.qa_loader import generate_qa_from_folder
                        qa_pairs = generate_qa_from_folder(folder_path, output_file, num_questions, sentence_length=sentence_length)
                        st.success(f"Successfully generated {len(qa_pairs)} QA pairs!")
                        # 自动加载生成的数据集
                        st.session_state.qa_dataset = get_qa_dataset_("custom", output_file)
                        cfg.dataset = "custom"
                        st.session_state.step = 2
                        st.rerun()
                except Exception as e:
                    st.error(f"Error generating QA dataset: {str(e)}")

        if st.button("Load Dataset"):
            if uploaded_file is not None:
                try:
                    st.session_state.step = 2
                    with st.spinner("Loading Dataset..."):
                        st.session_state.qa_dataset = get_qa_dataset_("custom", uploaded_file)
                        cfg.dataset = "custom"
                    st.rerun()
                except Exception as e:
                    st.error(f"Error loading dataset: {str(e)}")
            else:
                with st.spinner("Loading Dataset..."):
                    st.session_state.qa_dataset = get_qa_dataset_(cfg.dataset)
                st.session_state.step = 2
                st.rerun()

    if st.session_state.step == 2:
        st.header("Configure your RAG Index")
        st.subheader("Settings")

        cfg.llm = st.selectbox("LLM", options=LLM_OPTIONS, index=LLM_OPTIONS.index(cfg.llm) if cfg.llm in LLM_OPTIONS else 0)
        if cfg.llm == "ollama":
            cfg.ollama_model = st.text_input("Your Ollama Model", value=cfg.ollama_model)
        elif cfg.llm == 'huggingface':
            cfg.huggingface_model = st.selectbox("HuggingFace Model", options=HF_MODEL_OPTIONS, index=HF_MODEL_OPTIONS.index(cfg.huggingface_model) if cfg.huggingface_model in HF_MODEL_OPTIONS else 0)
            cfg.auth_token = st.text_input("Your Auth Token", value=cfg.auth_token)
        elif cfg.llm == 'openai':
            cfg.api_key = st.text_input("API Key", value=cfg.api_key, type="password")
            cfg.api_base = st.text_input("API Base", value=cfg.api_base)
            cfg.api_name = st.text_input("Model Name", value=cfg.api_name)

        st.markdown("---")
        cfg.embeddings = st.selectbox("Embeddings", options=EMBEDDING_OPTIONS, index=EMBEDDING_OPTIONS.index(cfg.embeddings) if cfg.embeddings in EMBEDDING_OPTIONS else 0)
        cfg.split_type = st.selectbox("Split Type", options=SPLIT_TYPE_OPTIONS, index=SPLIT_TYPE_OPTIONS.index(cfg.split_type))
        cfg.chunk_size = st.number_input("Chunk Size", min_value=1, value=cfg.chunk_size, step=1)
        # cfg.source_dir = st.text_input("Source Directory", value=cfg.source_dir)
        cfg.persist_dir = st.text_input("Persist Directory", value=cfg.persist_dir)

        # 返回或者继续
        c1,c_,c2 = st.columns([1,4,1])
        with c1:
            if st.button("Back"):
                st.session_state.step = 1
                st.rerun()
        with c2:
            if st.button("Build Index"):
                st.session_state.step = 3
                with st.spinner("Building Index..."):
                    st.session_state.index, st.session_state.hierarchical_storage_context = get_index()
                st.rerun()
            # Evaluation Metrics Selection

    if st.session_state.step == 3:
        st.header("Configure your RAG Query Engine")
        cfg.retriever = st.selectbox("Advanced Retriever", options=RETRIEVER_OPTIONS, index=RETRIEVER_OPTIONS.index(
            cfg.retriever) if cfg.retriever in RETRIEVER_OPTIONS else 0)
        cfg.retriever_mode = st.selectbox("Retriever Mode", options=[0, 1], index=cfg.retriever_mode)
        cfg.query_transform = st.selectbox("Pre-retriever", options=QUERY_TRANSFORM_OPTIONS,
                                           index=QUERY_TRANSFORM_OPTIONS.index(
                                               cfg.query_transform) if cfg.query_transform in QUERY_TRANSFORM_OPTIONS else 0)
        cfg.postprocess_rerank = st.selectbox("Post-process", options=POSTPROCESS_RERANK_OPTIONS,
                                              index=POSTPROCESS_RERANK_OPTIONS.index(
                                                  cfg.postprocess_rerank) if cfg.postprocess_rerank in POSTPROCESS_RERANK_OPTIONS else 0)
        
        cfg.text_qa_template_str = st.text_area("Text QA Template", value=cfg.text_qa_template_str)
        cfg.refine_template_str = st.text_area("Refine Template", value=cfg.refine_template_str)

        c1, c_, c2 = st.columns([1, 4, 1])
        with c1:
            if st.button("Back"):
                st.session_state.step = 2
                st.rerun()
        with c2:
            if st.button("Build Query Engine"):
                st.session_state.step = 4
                with st.spinner("Building Query Engine..."):
                    st.session_state.query_engine = get_query_engine()
                st.rerun()
    if st.session_state.step == 4:
        st.header("Evaluate your RAG Model with single question")
        prompt = st.text_input('Input your question here')
        if st.button("Evaluate Your Question"):
            response = transform_and_query(prompt, cfg, st.session_state.query_engine)
            st.write(response.response)

            # Display source text
            with st.expander('Source Text'):
                st.write(response.get_formatted_sources(length=1024))

        c1, c_, c2 = st.columns([1, 4, 1])
        with c1:
            if st.button("Back"):
                st.session_state.step = 3
                st.rerun()

        with c2:
            if st.button("Evaluate Your Dataset"):
                st.session_state.step = 5
                st.rerun()

    if st.session_state.step == 5:
        st.header("Evaluate your RAG Model with your dataset")
        
        # 确保默认选中的 metrics 按 AVAILABLE_METRICS 的顺序排列
        sorted_default_metrics = [metric for metric in AVAILABLE_METRICS if metric in cfg.metrics]
        sorted_frontend_default_metrics = [METRIC_DISPLAY_MAP.get(metric, metric) for metric in sorted_default_metrics]
        
        # 用户在前端选择评测指标
        selected_frontend_metrics = st.multiselect(
            "Evaluation Metrics",
            options=FRONTEND_AVAILABLE_METRICS,
            default=sorted_frontend_default_metrics
        )
        
        # 将前端选择的显示名称映射回后端的真实名称
        selected_backend_metrics = [DISPLAY_TO_BACKEND_METRIC_MAP.get(metric, metric) for metric in selected_frontend_metrics]
        cfg.metrics = selected_backend_metrics

        
        # 其他输入
        cfg.test_init_total_number_documents = st.number_input(
            "Total number of documents to evaluate", 
            min_value=1, 
            value=cfg.test_init_total_number_documents, 
            step=1
        )

        c1, c_, c2 = st.columns([1, 4, 1])
        with c1:
            if st.button("Back"):
                st.session_state.step = 4
                st.rerun()

        with c2:
            start_evaluation = st.button("Start Evaluation")
        if start_evaluation:
            all_num = 0
            metrics = cfg.metrics.copy()
            evaluateResults = EvaluationResult(metrics=metrics)
            evalAgent = EvalModelAgent(cfg)
            if cfg.experiment_1:
                if len(st.session_state.qa_dataset) < cfg.test_init_total_number_documents:
                    warnings.filterwarnings('default')
                    warnings.warn("使用的数据集长度大于数据集本身的最大长度，请修改。 本轮代码无法运行", UserWarning)
            else:
                cfg.test_init_total_number_documents = cfg.n
            for question, expected_answer, golden_context, golden_context_ids in zip(
                    st.session_state.qa_dataset['test_data']['question'][:cfg.test_init_total_number_documents],
                    st.session_state.qa_dataset['test_data']['expected_answer'][:cfg.test_init_total_number_documents],
                    st.session_state.qa_dataset['test_data']['golden_context'][:cfg.test_init_total_number_documents],
                    st.session_state.qa_dataset['test_data']['golden_context_ids'][:cfg.test_init_total_number_documents]
            ):
                response = transform_and_query(question, cfg, st.session_state.query_engine)
                # 返回node节点
                retrieval_ids = []
                retrieval_context = []
                for source_node in response.source_nodes:
                    retrieval_ids.append(source_node.metadata['id'])
                    retrieval_context.append(source_node.get_content())
                actual_response = response.response
                eval_result = evaluating(
                    question, response, actual_response, retrieval_context, retrieval_ids,
                    expected_answer, golden_context, golden_context_ids, evaluateResults.metrics,
                    evalAgent
                )
                with st.expander(question):
                    st.markdown("### Answer")
                    st.markdown(response.response)
                    st.markdown('### Retrieval context')
                    st.markdown('\n\n'.join(retrieval_context))
                    st.markdown('### Expected answer')
                    st.markdown(expected_answer)
                    st.markdown('### Golden context')
                    st.markdown('\n\n'.join(golden_context))

                print(eval_result)

                evaluateResults.add(eval_result)
                all_num += 1
                st.markdown(evaluateResults.get_results_str())

            st.success("Evaluation complete!")
            st.session_state.evaluation_results = evaluateResults

    # if st.session_state.step == 5:
    #     st.header("Evaluation Results")
    #     if 'evaluation_results' in st.session_state:
    #         results = st.session_state.evaluation_results
    #         display_results(results)

def display_results(results: EvaluationResult):
    # Display summary statistics
    st.subheader("Summary Statistics")
    summary = pd.DataFrame(results.get_summary(), index=[0])
    st.table(summary)

    # Display detailed metrics
    st.subheader("Detailed Metrics")
    metrics_df = pd.DataFrame(results.get_all_metrics())
    st.dataframe(metrics_df)

    # Visualize metrics
    st.subheader("Metric Visualization")
    metric_to_plot = st.selectbox("Select a metric to visualize", options=results.metrics)
    fig = px.box(metrics_df, y=metric_to_plot, title=f"Distribution of {metric_to_plot}")
    st.plotly_chart(fig)

    # Display sample evaluations
    st.subheader("Sample Evaluations")
    num_samples = st.slider("Number of samples to display", min_value=1, max_value=len(results.evaluations), value=5)
    for i, eval_result in enumerate(results.evaluations[:num_samples]):
        st.write(f"Sample {i+1}")
        st.json(eval_result)

if __name__ == "__main__":
    main()



import argparse

import deepeval.api
import torch
from deepeval.models import GPTModel
from langchain_openai import ChatOpenAI
from llama_index.llms.huggingface import HuggingFaceLLM
from llama_index.llms.openai import OpenAI
from transformers import AutoModelForCausalLM, AutoTokenizer
from deepeval.models.base_model import DeepEvalBaseLLM
from uptrain import Settings
from .DeepEvalLocalModel import DeepEvalLocalModel

load_tokenizer = []
llm_args = {"context_window": 4096, "max_new_tokens": 256,
                             "generate_kwargs": {"temperature": 0.7, "top_k": 50, "top_p": 0.95}}
def qwen_completion_to_prompt(completion):
    tokenizer = load_tokenizer[0]
    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": completion}
    ]
    return tokenizer.apply_chat_template(
        messages,
        tokenize=False,
        add_generation_prompt=True
    )
class EvalModelAgent():
    def __init__(self, args):
        self.args = args
        llamaIndex_LocalmodelName = self.args.llamaIndexEvaluateModel
        deepEval_LocalModelName = self.args.deepEvalEvaluateModel
        uptrain_LocalModelName = self.args.upTrainEvaluateModel
        api_name = self.args.api_name
        api_key = self.args.api_key
        api_base = self.args.api_base
        print("EvalModelName:")
        print(api_name)
        print("EvalModelAPI:")
        print(api_key)
        if api_name == "":
            self._llama_model = AutoModelForCausalLM.from_pretrained(llamaIndex_LocalmodelName,
                                                     torch_dtype=torch.float16,
                                                     device_map="auto").eval()
            self._llama_tokenizer = AutoTokenizer.from_pretrained(llamaIndex_LocalmodelName)
            load_tokenizer.append(self._llama_tokenizer)
            self.llamaModel = HuggingFaceLLM(context_window=llm_args["context_window"],
                              max_new_tokens=llm_args["max_new_tokens"],
                              completion_to_prompt=qwen_completion_to_prompt,
                              generate_kwargs=llm_args["generate_kwargs"],
                              model=self._llama_model,
                              tokenizer=self._llama_tokenizer,
                              device_map="cuda:0",)
        else:
            self.llamaModel = OpenAI(api_key=api_key, api_base=api_base,
                      model=api_name)
        if api_name == "":
            if deepEval_LocalModelName == llamaIndex_LocalmodelName:
                self._deepEval_model = self._llama_model
                self._deepEval_tokenizer = self._llama_tokenizer
            else:
                self._deepEval_model = AutoModelForCausalLM.from_pretrained(deepEval_LocalModelName,
                                                     torch_dtype=torch.float16,
                                                     device_map="auto").eval()
                self._deepEval_tokenizer = AutoTokenizer.from_pretrained(deepEval_LocalModelName)
            self.deepEvalModel = DeepEvalLocalModel(model=self._deepEval_model,
                                                    tokenizer=self._deepEval_tokenizer)
        else:
            # 不再有效了
            # deepeval.api.API_BASE_URL = 'https://uiuiapi.com/v1'
            # self.deepEvalModel = api_name

            self._deepEval_model = ChatOpenAI(openai_api_key=api_key, openai_api_base=api_base,
                      model_name=api_name)
            self.deepEvalModel = DeepEvalLocalModel(model=self._deepEval_model,
                                                    tokenizer="")
        if api_name == "":
            self.uptrainSetting = Settings(model="ollama/"+uptrain_LocalModelName)
        else:
            self.uptrainSetting = Settings(
                    model=api_name,
                    openai_api_key=api_key,
                    base_url=api_base,
                )


from jury import Jury
import jury
import evaluate

NLG_EVALUATION_METRICS = [
    "chrf", "meteor", "wer", "cer", "chrf_pp", "mauve", "perplexity",
    "rouge_rouge1", "rouge_rouge2", "rouge_rougeL", "rouge_rougeLsum"
]


def NLGEvaluate(questions, actual_responses, golden_contexts, golden_context_ids, metrics):
    # omit_metrics = []
    # for metric in NLG_EVALUATION_METRICS:
    #     if metric not in metrics:
    #         omit_metrics.append(metric)

    # n = NLGEval(metrics_to_omit=omit_metrics)
    references = []
    for context in golden_contexts:
        references.append(str(context))
    if type(actual_responses) == list:
        predictions = [str(response) for response in actual_responses]
    elif type(actual_responses) == str:
        predictions = [actual_responses]

    # Individual Metrics
    # scores = n.compute_individual_metrics(ref=reference, hyp=hypothesis)
    scorer = Jury(metrics=["chrf", "meteor", "rouge", "wer", "cer"])
    scores = {}
    # chrf++
    chrf_plus = evaluate.load("chrf")
    score = chrf_plus.compute(predictions=predictions, references=[references], word_order=2)
    scores["chrf_pp"] = score["score"]
    # MAUVE 一个 predictions 只能对应一个references
    mauve = evaluate.load('mauve')
    score = mauve.compute(predictions=predictions, references=[''.join(references)])
    scores["mauve"] = score.mauve
    # perplexity:model id are needed
    perplexity = jury.load_metric("perplexity")
    score = perplexity.compute(predictions=predictions, references=references, model_id="openai-community/gpt2")
    scores["perplexity"] = score["mean_perplexity"]
    #
    score = scorer(predictions=predictions, references=[references])
    scores["chrf"] = score["chrf"]["score"]
    scores["meteor"] = score["meteor"]["score"]
    # 'rouge1': 0.6666666666666665, 'rouge2': 0.5714285714285715, 'rougeL': 0.6666666666666665, 'rougeLsum': 0.6666666666666665
    scores["rouge_rouge1"] = score["rouge"]["rouge1"]
    scores["rouge_rouge2"] = score["rouge"]["rouge2"]
    scores["rouge_rougeL"] = score["rouge"]["rougeL"]
    scores["rouge_rougeLsum"] = score["rouge"]["rougeLsum"]
    scores["wer"] = score["wer"]["score"]
    scores["cer"] = score["cer"]["score"]
    return scores



from llama_index.core.postprocessor import LongContextReorder
from llama_index.postprocessor.colbert_rerank import ColbertRerank
from llama_index.postprocessor.cohere_rerank import CohereRerank
from llama_index.postprocessor.flag_embedding_reranker import FlagEmbeddingReranker

# !pip install llama-index-postprocessor-colbert-rerank
# !pip install llama-index-postprocessor-cohere-rerank

def get_postprocessor(cfg):
    # postprocess rerank, available: long_context_reorder, colbertv2_rerank, cohere_rerank, bge-reranker-base
    if cfg.postprocess_rerank == 'long_context_reorder':
        return LongContextReorder()
    elif cfg.postprocess_rerank == 'colbertv2_rerank':
        return ColbertRerank()
    elif cfg.postprocess_rerank == 'cohere_rerank':
        return CohereRerank()
    elif cfg.postprocess_rerank == 'bge-reranker-base':
        return FlagEmbeddingReranker(model="BAAI/bge-reranker-base")
    else:
        raise Exception("postprocess_rerank not supported: %s" % cfg.postprocess_rerank)


import numpy as np
from llama_index.core import PromptTemplate

LLAMA_CUSTOM_FAITHFULNESS_TEMPLATE = PromptTemplate(
    "Please tell if the context supports the given information related to the question.\n"
    "You need to answer with either YES or NO.\n"
    "Answer YES if the context supports the information related to the question, even if most of the context is unrelated. \n"
    "See the examples below.\n"
    "Question: Is apple pie usually double-crusted?\n"
    "Information: Apple pie is generally double-crusted.\n"
    "Context: An apple pie is a fruit pie in which the principal filling ingredient is apples. \n"
    "Apple pie is often served with whipped cream, ice cream ('apple pie à la mode'), custard or cheddar cheese.\n"
    "It is generally double-crusted, with pastry both above and below the filling; the upper crust may be solid or latticed (woven of crosswise strips).\n"
    "Answer: YES\n"
    "Question: Does apple pie usually taste bad?\n"
    "Information: Apple pies tastes bad.\n"
    "Context: An apple pie is a fruit pie in which the principal filling ingredient is apples. \n"
    "Apple pie is often served with whipped cream, ice cream ('apple pie à la mode'), custard or cheddar cheese.\n"
    "It is generally double-crusted, with pastry both above and below the filling; the upper crust may be solid or latticed (woven of crosswise strips).\n"
    "Answer: NO\n"
    "Question and Information: {query_str}\n"
    "Context: {context_str}\n"
    "Answer: "
)


class EvaluationResult_TRT:
    def __init__(self, metrics=None):
        self.results = {
            "n": 0,
            "F1": 0.0,
            "em": 0.0,
            "mrr": 0.0,
            "hit1": 0.0,
            "hit10": 0.0,
            "MAP": 0.0,
            "NDCG": 0.0,
            "DCG": 0.0,
            "IDCG": 0.0,
        }
        # 初始化用于评估的指标，每一个指标都是一个包含score和count的字典
        evaluation_metrics = []
        self.evaluationName = evaluation_metrics
        evaluation_metrics_rev = []
        for i in evaluation_metrics:
            evaluation_metrics_rev.append(i + "_rev")
        self.metrics_results = {}
        for metric in evaluation_metrics:
            self.metrics_results[metric] = {"score": 0, "count": 0}
            self.metrics_results[metric + "_rev"] = {"score": 0, "count": 0}

        if metrics is None:
            metrics = []
        metrics.append("n")
        metrics.append("F1")
        metrics.append("em")
        metrics.append("mrr")
        metrics.append("hit1")
        metrics.append("hit10")
        metrics.append("MAP")
        metrics.append("NDCG")
        metrics.append("DCG")
        metrics.append("IDCG")

        self.metrics = metrics

    def add(self, evaluate_result):
        for key in self.results.keys():
            if key in self.metrics:
                self.results[key] += evaluate_result.results[key]
        for key in self.metrics_results.keys():
            if key in self.metrics:
                if evaluate_result.metrics_results[key]["score"] != None:
                    self.metrics_results[key]["score"] += evaluate_result.metrics_results[key]["score"]
                    self.metrics_results[key]["count"] += evaluate_result.metrics_results[key]["count"]
        self.results["n"] += 1

    def print_results(self):
        for key, value in self.results.items():
            if key in self.metrics:
                if key == 'n':
                    print(f"{key}: {value}")
                else:
                    print(f"{key}: {value / self.results['n']}")
        for key, value in self.metrics_results.items():
            if key in self.metrics:
                if value['count'] == 0:
                    print(f"{key}: 0, valid number : {value['count']}")
                else:
                    print(f"{key}: {value['score'] / value['count']}, valid number : {value['count']}")

    def print_results_to_path(self, path, config, sample_arr):
        print("save data to " + path)
        f = open(path, 'a')
        f.writelines("\n")
        f.writelines("=========================== begin ==================\n")
        f.writelines("database: " + config.dataset + "\n")
        f.writelines("sample_arr: " + str(sample_arr) + "\n")
        for key, value in self.results.items():
            if key in self.metrics:
                if key == 'n':
                    f.writelines(f"{key}: {value}\n")
                else:
                    f.writelines(f"{key}: {value / self.results['n']}\n")
        for key, value in self.metrics_results.items():
            if key in self.metrics:
                if value['count'] == 0:
                    f.writelines(f"{key}: 0, valid number : {value['count']}\n")
                else:
                    f.writelines(f"{key}: {value['score'] / value['count']}, valid number : {value['count']}\n")


# response evaluate
def evaluating_TRT(retrieval_ids, golden_context_ids):
    # 创建一个新类，主要是用来记录各个指标有效的个数以及得分
    eval_result = EvaluationResult_TRT()
    # region 常规指标
    eval_result.results["F1"] = F1(retrieval_ids, golden_context_ids)
    eval_result.results["em"] = Em(retrieval_ids, golden_context_ids)
    eval_result.results["mrr"] = Mrr(retrieval_ids, golden_context_ids)
    eval_result.results["hit1"] = Hit(retrieval_ids, golden_context_ids[0:1])
    eval_result.results["hit10"] = Hit(retrieval_ids, golden_context_ids[0:10])
    eval_result.results["MAP"] = MAP(retrieval_ids, golden_context_ids)
    eval_result.results["NDCG"] = NDCG(retrieval_ids, golden_context_ids)
    eval_result.results["DCG"] = DCG(retrieval_ids, golden_context_ids)
    eval_result.results["IDCG"] = IDCG(retrieval_ids, golden_context_ids)
    # endregion
    # 由于upTrain可以一次计算多个指标，所以这个变量之后会从upTrain的多个指标

    # endregion
    return eval_result


# region commonly used indicators
def Mrr(retrieved_ids, expected_ids):
    if retrieved_ids is None or expected_ids is None:
        raise ValueError("Retrieved ids and expected ids must be provided")
    for i, id in enumerate(retrieved_ids):
        if id in expected_ids:
            score = 1.0 / (i + 1)
            return score
    return 0.0


def Hit(retrieved_ids, expected_ids):
    if retrieved_ids is None or expected_ids is None:
        raise ValueError("Retrieved ids and expected ids must be provided")
    is_hit = any(id in expected_ids for id in retrieved_ids)
    score = 1.0 if is_hit else 0.0
    return score


def F1(retrieved_ids, expected_ids):
    # 转换为集合进行交集和差集运算
    retrieved_set = set(retrieved_ids)
    expected_set = set(expected_ids)

    # 计算TP、FP和FN
    TP = len(retrieved_set & expected_set)  # 预测为正且实际为正的
    FP = len(retrieved_set - expected_set)  # 预测为正但实际为负的
    FN = len(expected_set - retrieved_set)  # 实际为正但预测为负的

    # 计算精确率（Precision）和召回率（Recall）
    Precision = TP / (TP + FP) if (TP + FP) > 0 else 0
    Recall = TP / (TP + FN) if (TP + FN) > 0 else 0

    # 计算F1分数
    F1_score = 2 * Precision * Recall / (Precision + Recall) if (Precision + Recall) > 0 else 0

    return F1_score


def Em(retrieved_ids, expected_ids):
    retrieved_ids.sort()
    expected_ids.sort()
    if expected_ids == retrieved_ids:
        return 1
    return 0


def MAP(retrieved_ids, expected_ids):
    if retrieved_ids is None or expected_ids is None:
        raise ValueError("Retrieved ids and expected ids must be provided")
    if len(retrieved_ids) == 0:
        return 0.0
    score = 0.0
    for i, id in enumerate(expected_ids):
        if id in retrieved_ids:
            score += (i + 1) / (retrieved_ids.index(id) + 1)
    return score / len(expected_ids)


def DCG(retrieved_ids, expected_ids):
    if retrieved_ids is None or expected_ids is None:
        raise ValueError("Retrieved ids and expected ids must be provided")
    score = 0.0
    for i, id in enumerate(retrieved_ids):
        if id in expected_ids:
            # index starts at 0, so add 1 to the index
            score += (1) / np.log2((i + 1) + 1)

    return score


def IDCG(retreived_ids, expected_ids):
    temp_1 = []
    temp_2 = []
    for a in retreived_ids:
        if a in expected_ids:
            temp_1.append(a)
        else:
            temp_2.append(a)
    temp_1.append(temp_2)
    idcg = DCG(temp_1, expected_ids)
    return idcg


def NDCG(retrieved_ids, expected_ids):
    dcg = DCG(retrieved_ids, expected_ids)
    idcg = IDCG(retrieved_ids, expected_ids)
    if idcg == 0:
        return 0
    return dcg / idcg
# endregion
